from __future__ import annotations

import zlib
from collections.abc import Sequence
from typing import Any

from .. import frames
from ..exceptions import (
    DuplicateParameter,
    InvalidParameterName,
    InvalidParameterValue,
    NegotiationError,
    PayloadTooBig,
    ProtocolError,
)
from ..typing import ExtensionName, ExtensionParameter
from .base import ClientExtensionFactory, Extension, ServerExtensionFactory


__all__ = [
    "PerMessageDeflate",
    "ClientPerMessageDeflateFactory",
    "enable_client_permessage_deflate",
    "ServerPerMessageDeflateFactory",
    "enable_server_permessage_deflate",
]

_EMPTY_UNCOMPRESSED_BLOCK = b"\x00\x00\xff\xff"

_MAX_WINDOW_BITS_VALUES = [str(bits) for bits in range(8, 16)]


class PerMessageDeflate(Extension):
    """
    Per-Message Deflate extension.

    """

    name = ExtensionName("permessage-deflate")

    def __init__(
        self,
        remote_no_context_takeover: bool,
        local_no_context_takeover: bool,
        remote_max_window_bits: int,
        local_max_window_bits: int,
        compress_settings: dict[Any, Any] | None = None,
    ) -> None:
        """
        Configure the Per-Message Deflate extension.

        """
        if compress_settings is None:
            compress_settings = {}

        assert remote_no_context_takeover in [False, True]
        assert local_no_context_takeover in [False, True]
        assert 8 <= remote_max_window_bits <= 15
        assert 8 <= local_max_window_bits <= 15
        assert "wbits" not in compress_settings

        self.remote_no_context_takeover = remote_no_context_takeover
        self.local_no_context_takeover = local_no_context_takeover
        self.remote_max_window_bits = remote_max_window_bits
        self.local_max_window_bits = local_max_window_bits
        self.compress_settings = compress_settings

        if not self.remote_no_context_takeover:
            self.decoder = zlib.decompressobj(wbits=-self.remote_max_window_bits)

        if not self.local_no_context_takeover:
            self.encoder = zlib.compressobj(
                wbits=-self.local_max_window_bits,
                **self.compress_settings,
            )

        # To handle continuation frames properly, we must keep track of
        # whether that initial frame was encoded.
        self.decode_cont_data = False
        # There's no need for self.encode_cont_data because we always encode
        # outgoing frames, so it would always be True.

    def __repr__(self) -> str:
        return (
            f"PerMessageDeflate("
            f"remote_no_context_takeover={self.remote_no_context_takeover}, "
            f"local_no_context_takeover={self.local_no_context_takeover}, "
            f"remote_max_window_bits={self.remote_max_window_bits}, "
            f"local_max_window_bits={self.local_max_window_bits})"
        )

    def decode(
        self,
        frame: frames.Frame,
        *,
        max_size: int | None = None,
    ) -> frames.Frame:
        """
        Decode an incoming frame.

        """
        # Skip control frames.
        if frame.opcode in frames.CTRL_OPCODES:
            return frame

        # Handle continuation data frames:
        # - skip if the message isn't encoded
        # - reset "decode continuation data" flag if it's a final frame
        if frame.opcode is frames.OP_CONT:
            if not self.decode_cont_data:
                return frame
            if frame.fin:
                self.decode_cont_data = False

        # Handle text and binary data frames:
        # - skip if the message isn't encoded
        # - unset the rsv1 flag on the first frame of a compressed message
        # - set "decode continuation data" flag if it's a non-final frame
        else:
            if not frame.rsv1:
                return frame
            if not frame.fin:
                self.decode_cont_data = True

            # Re-initialize per-message decoder.
            if self.remote_no_context_takeover:
                self.decoder = zlib.decompressobj(wbits=-self.remote_max_window_bits)

        # Uncompress data. Protect against zip bombs by preventing zlib from
        # decompressing more than max_length bytes (except when the limit is
        # disabled with max_size = None).
        if frame.fin and len(frame.data) < 2044:
            # Profiling shows that appending four bytes, which makes a copy, is
            # faster than calling decompress() again when data is less than 2kB.
            data = bytes(frame.data) + _EMPTY_UNCOMPRESSED_BLOCK
        else:
            data = frame.data
        max_length = 0 if max_size is None else max_size
        try:
            data = self.decoder.decompress(data, max_length)
            if self.decoder.unconsumed_tail:
                assert max_size is not None  # help mypy
                raise PayloadTooBig(None, max_size)
            if frame.fin and len(frame.data) >= 2044:
                # This cannot generate additional data.
                self.decoder.decompress(_EMPTY_UNCOMPRESSED_BLOCK)
        except zlib.error as exc:
            raise ProtocolError("decompression failed") from exc

        # Allow garbage collection of the decoder if it won't be reused.
        if frame.fin and self.remote_no_context_takeover:
            del self.decoder

        return frames.Frame(
            frame.opcode,
            data,
            frame.fin,
            # Unset the rsv1 flag on the first frame of a compressed message.
            False,
            frame.rsv2,
            frame.rsv3,
        )

    def encode(self, frame: frames.Frame) -> frames.Frame:
        """
        Encode an outgoing frame.

        """
        # Skip control frames.
        if frame.opcode in frames.CTRL_OPCODES:
            return frame

        # Since we always encode messages, there's no "encode continuation
        # data" flag similar to "decode continuation data" at this time.

        if frame.opcode is not frames.OP_CONT:
            # Re-initialize per-message decoder.
            if self.local_no_context_takeover:
                self.encoder = zlib.compressobj(
                    wbits=-self.local_max_window_bits,
                    **self.compress_settings,
                )

        # Compress data.
        data = self.encoder.compress(frame.data) + self.encoder.flush(zlib.Z_SYNC_FLUSH)
        if frame.fin:
            # Sync flush generates between 5 or 6 bytes, ending with the bytes
            # 0x00 0x00 0xff 0xff, which must be removed.
            assert data[-4:] == _EMPTY_UNCOMPRESSED_BLOCK
            # Making a copy is faster than memoryview(a)[:-4] until 2kB.
            if len(data) < 2048:
                data = data[:-4]
            else:
                data = memoryview(data)[:-4]

        # Allow garbage collection of the encoder if it won't be reused.
        if frame.fin and self.local_no_context_takeover:
            del self.encoder

        return frames.Frame(
            frame.opcode,
            data,
            frame.fin,
            # Set the rsv1 flag on the first frame of a compressed message.
            frame.opcode is not frames.OP_CONT,
            frame.rsv2,
            frame.rsv3,
        )


def _build_parameters(
    server_no_context_takeover: bool,
    client_no_context_takeover: bool,
    server_max_window_bits: int | None,
    client_max_window_bits: int | bool | None,
) -> list[ExtensionParameter]:
    """
    Build a list of ``(name, value)`` pairs for some compression parameters.

    """
    params: list[ExtensionParameter] = []
    if server_no_context_takeover:
        params.append(("server_no_context_takeover", None))
    if client_no_context_takeover:
        params.append(("client_no_context_takeover", None))
    if server_max_window_bits:
        params.append(("server_max_window_bits", str(server_max_window_bits)))
    if client_max_window_bits is True:  # only in handshake requests
        params.append(("client_max_window_bits", None))
    elif client_max_window_bits:
        params.append(("client_max_window_bits", str(client_max_window_bits)))
    return params


def _extract_parameters(
    params: Sequence[ExtensionParameter], *, is_server: bool
) -> tuple[bool, bool, int | None, int | bool | None]:
    """
    Extract compression parameters from a list of ``(name, value)`` pairs.

    If ``is_server`` is :obj:`True`, ``client_max_window_bits`` may be
    provided without a value. This is only allowed in handshake requests.

    """
    server_no_context_takeover: bool = False
    client_no_context_takeover: bool = False
    server_max_window_bits: int | None = None
    client_max_window_bits: int | bool | None = None

    for name, value in params:
        if name == "server_no_context_takeover":
            if server_no_context_takeover:
                raise DuplicateParameter(name)
            if value is None:
                server_no_context_takeover = True
            else:
                raise InvalidParameterValue(name, value)

        elif name == "client_no_context_takeover":
            if client_no_context_takeover:
                raise DuplicateParameter(name)
            if value is None:
                client_no_context_takeover = True
            else:
                raise InvalidParameterValue(name, value)

        elif name == "server_max_window_bits":
            if server_max_window_bits is not None:
                raise DuplicateParameter(name)
            if value in _MAX_WINDOW_BITS_VALUES:
                server_max_window_bits = int(value)
            else:
                raise InvalidParameterValue(name, value)

        elif name == "client_max_window_bits":
            if client_max_window_bits is not None:
                raise DuplicateParameter(name)
            if is_server and value is None:  # only in handshake requests
                client_max_window_bits = True
            elif value in _MAX_WINDOW_BITS_VALUES:
                client_max_window_bits = int(value)
            else:
                raise InvalidParameterValue(name, value)

        else:
            raise InvalidParameterName(name)

    return (
        server_no_context_takeover,
        client_no_context_takeover,
        server_max_window_bits,
        client_max_window_bits,
    )


class ClientPerMessageDeflateFactory(ClientExtensionFactory):
    """
    Client-side extension factory for the Per-Message Deflate extension.

    Parameters behave as described in `section 7.1 of RFC 7692`_.

    .. _section 7.1 of RFC 7692: https://datatracker.ietf.org/doc/html/rfc7692#section-7.1

    Set them to :obj:`True` to include them in the negotiation offer without a
    value or to an integer value to include them with this value.

    Args:
        server_no_context_takeover: Prevent server from using context takeover.
        client_no_context_takeover: Prevent client from using context takeover.
        server_max_window_bits: Maximum size of the server's LZ77 sliding window
            in bits, between 8 and 15.
        client_max_window_bits: Maximum size of the client's LZ77 sliding window
            in bits, between 8 and 15, or :obj:`True` to indicate support without
            setting a limit.
        compress_settings: Additional keyword arguments for :func:`zlib.compressobj`,
            excluding ``wbits``.

    """

    name = ExtensionName("permessage-deflate")

    def __init__(
        self,
        server_no_context_takeover: bool = False,
        client_no_context_takeover: bool = False,
        server_max_window_bits: int | None = None,
        client_max_window_bits: int | bool | None = True,
        compress_settings: dict[str, Any] | None = None,
    ) -> None:
        """
        Configure the Per-Message Deflate extension factory.

        """
        if not (server_max_window_bits is None or 8 <= server_max_window_bits <= 15):
            raise ValueError("server_max_window_bits must be between 8 and 15")
        if not (
            client_max_window_bits is None
            or client_max_window_bits is True
            or 8 <= client_max_window_bits <= 15
        ):
            raise ValueError("client_max_window_bits must be between 8 and 15")
        if compress_settings is not None and "wbits" in compress_settings:
            raise ValueError(
                "compress_settings must not include wbits, "
                "set client_max_window_bits instead"
            )

        self.server_no_context_takeover = server_no_context_takeover
        self.client_no_context_takeover = client_no_context_takeover
        self.server_max_window_bits = server_max_window_bits
        self.client_max_window_bits = client_max_window_bits
        self.compress_settings = compress_settings

    def get_request_params(self) -> list[ExtensionParameter]:
        """
        Build request parameters.

        """
        return _build_parameters(
            self.server_no_context_takeover,
            self.client_no_context_takeover,
            self.server_max_window_bits,
            self.client_max_window_bits,
        )

    def process_response_params(
        self,
        params: Sequence[ExtensionParameter],
        accepted_extensions: Sequence[Extension],
    ) -> PerMessageDeflate:
        """
        Process response parameters.

        Return an extension instance.

        """
        if any(other.name == self.name for other in accepted_extensions):
            raise NegotiationError(f"received duplicate {self.name}")

        # Request parameters are available in instance variables.

        # Load response parameters in local variables.
        (
            server_no_context_takeover,
            client_no_context_takeover,
            server_max_window_bits,
            client_max_window_bits,
        ) = _extract_parameters(params, is_server=False)

        # After comparing the request and the response, the final
        # configuration must be available in the local variables.

        # server_no_context_takeover
        #
        #   Req.    Resp.   Result
        #   ------  ------  --------------------------------------------------
        #   False   False   False
        #   False   True    True
        #   True    False   Error!
        #   True    True    True

        if self.server_no_context_takeover:
            if not server_no_context_takeover:
                raise NegotiationError("expected server_no_context_takeover")

        # client_no_context_takeover
        #
        #   Req.    Resp.   Result
        #   ------  ------  --------------------------------------------------
        #   False   False   False
        #   False   True    True
        #   True    False   True - must change value
        #   True    True    True

        if self.client_no_context_takeover:
            if not client_no_context_takeover:
                client_no_context_takeover = True

        # server_max_window_bits

        #   Req.    Resp.   Result
        #   ------  ------  --------------------------------------------------
        #   None    None    None
        #   None    8≤M≤15  M
        #   8≤N≤15  None    Error!
        #   8≤N≤15  8≤M≤N   M
        #   8≤N≤15  N<M≤15  Error!

        if self.server_max_window_bits is None:
            pass

        else:
            if server_max_window_bits is None:
                raise NegotiationError("expected server_max_window_bits")
            elif server_max_window_bits > self.server_max_window_bits:
                raise NegotiationError("unsupported server_max_window_bits")

        # client_max_window_bits

        #   Req.    Resp.   Result
        #   ------  ------  --------------------------------------------------
        #   None    None    None
        #   None    8≤M≤15  Error!
        #   True    None    None
        #   True    8≤M≤15  M
        #   8≤N≤15  None    N - must change value
        #   8≤N≤15  8≤M≤N   M
        #   8≤N≤15  N<M≤15  Error!

        if self.client_max_window_bits is None:
            if client_max_window_bits is not None:
                raise NegotiationError("unexpected client_max_window_bits")

        elif self.client_max_window_bits is True:
            pass

        else:
            if client_max_window_bits is None:
                client_max_window_bits = self.client_max_window_bits
            elif client_max_window_bits > self.client_max_window_bits:
                raise NegotiationError("unsupported client_max_window_bits")

        return PerMessageDeflate(
            server_no_context_takeover,  # remote_no_context_takeover
            client_no_context_takeover,  # local_no_context_takeover
            server_max_window_bits or 15,  # remote_max_window_bits
            client_max_window_bits or 15,  # local_max_window_bits
            self.compress_settings,
        )


def enable_client_permessage_deflate(
    extensions: Sequence[ClientExtensionFactory] | None,
) -> Sequence[ClientExtensionFactory]:
    """
    Enable Per-Message Deflate with default settings in client extensions.

    If the extension is already present, perhaps with non-default settings,
    the configuration isn't changed.

    """
    if extensions is None:
        extensions = []
    if not any(
        extension_factory.name == ClientPerMessageDeflateFactory.name
        for extension_factory in extensions
    ):
        extensions = list(extensions) + [
            ClientPerMessageDeflateFactory(
                compress_settings={"memLevel": 5},
            )
        ]
    return extensions


class ServerPerMessageDeflateFactory(ServerExtensionFactory):
    """
    Server-side extension factory for the Per-Message Deflate extension.

    Parameters behave as described in `section 7.1 of RFC 7692`_.

    .. _section 7.1 of RFC 7692: https://datatracker.ietf.org/doc/html/rfc7692#section-7.1

    Set them to :obj:`True` to include them in the negotiation offer without a
    value or to an integer value to include them with this value.

    Args:
        server_no_context_takeover: Prevent server from using context takeover.
        client_no_context_takeover: Prevent client from using context takeover.
        server_max_window_bits: Maximum size of the server's LZ77 sliding window
            in bits, between 8 and 15.
        client_max_window_bits: Maximum size of the client's LZ77 sliding window
            in bits, between 8 and 15.
        compress_settings: Additional keyword arguments for :func:`zlib.compressobj`,
            excluding ``wbits``.
        require_client_max_window_bits: Do not enable compression at all if
            client doesn't advertise support for ``client_max_window_bits``;
            the default behavior is to enable compression without enforcing
            ``client_max_window_bits``.

    """

    name = ExtensionName("permessage-deflate")

    def __init__(
        self,
        server_no_context_takeover: bool = False,
        client_no_context_takeover: bool = False,
        server_max_window_bits: int | None = None,
        client_max_window_bits: int | None = None,
        compress_settings: dict[str, Any] | None = None,
        require_client_max_window_bits: bool = False,
    ) -> None:
        """
        Configure the Per-Message Deflate extension factory.

        """
        if not (server_max_window_bits is None or 8 <= server_max_window_bits <= 15):
            raise ValueError("server_max_window_bits must be between 8 and 15")
        if not (client_max_window_bits is None or 8 <= client_max_window_bits <= 15):
            raise ValueError("client_max_window_bits must be between 8 and 15")
        if compress_settings is not None and "wbits" in compress_settings:
            raise ValueError(
                "compress_settings must not include wbits, "
                "set server_max_window_bits instead"
            )
        if client_max_window_bits is None and require_client_max_window_bits:
            raise ValueError(
                "require_client_max_window_bits is enabled, "
                "but client_max_window_bits isn't configured"
            )

        self.server_no_context_takeover = server_no_context_takeover
        self.client_no_context_takeover = client_no_context_takeover
        self.server_max_window_bits = server_max_window_bits
        self.client_max_window_bits = client_max_window_bits
        self.compress_settings = compress_settings
        self.require_client_max_window_bits = require_client_max_window_bits

    def process_request_params(
        self,
        params: Sequence[ExtensionParameter],
        accepted_extensions: Sequence[Extension],
    ) -> tuple[list[ExtensionParameter], PerMessageDeflate]:
        """
        Process request parameters.

        Return response params and an extension instance.

        """
        if any(other.name == self.name for other in accepted_extensions):
            raise NegotiationError(f"skipped duplicate {self.name}")

        # Load request parameters in local variables.
        (
            server_no_context_takeover,
            client_no_context_takeover,
            server_max_window_bits,
            client_max_window_bits,
        ) = _extract_parameters(params, is_server=True)

        # Configuration parameters are available in instance variables.

        # After comparing the request and the configuration, the response must
        # be available in the local variables.

        # server_no_context_takeover
        #
        #   Config  Req.    Resp.
        #   ------  ------  --------------------------------------------------
        #   False   False   False
        #   False   True    True
        #   True    False   True - must change value to True
        #   True    True    True

        if self.server_no_context_takeover:
            if not server_no_context_takeover:
                server_no_context_takeover = True

        # client_no_context_takeover
        #
        #   Config  Req.    Resp.
        #   ------  ------  --------------------------------------------------
        #   False   False   False
        #   False   True    True (or False)
        #   True    False   True - must change value to True
        #   True    True    True (or False)

        if self.client_no_context_takeover:
            if not client_no_context_takeover:
                client_no_context_takeover = True

        # server_max_window_bits

        #   Config  Req.    Resp.
        #   ------  ------  --------------------------------------------------
        #   None    None    None
        #   None    8≤M≤15  M
        #   8≤N≤15  None    N - must change value
        #   8≤N≤15  8≤M≤N   M
        #   8≤N≤15  N<M≤15  N - must change value

        if self.server_max_window_bits is None:
            pass

        else:
            if server_max_window_bits is None:
                server_max_window_bits = self.server_max_window_bits
            elif server_max_window_bits > self.server_max_window_bits:
                server_max_window_bits = self.server_max_window_bits

        # client_max_window_bits

        #   Config  Req.    Resp.
        #   ------  ------  --------------------------------------------------
        #   None    None    None
        #   None    True    None - must change value
        #   None    8≤M≤15  M (or None)
        #   8≤N≤15  None    None or Error!
        #   8≤N≤15  True    N - must change value
        #   8≤N≤15  8≤M≤N   M (or None)
        #   8≤N≤15  N<M≤15  N

        if self.client_max_window_bits is None:
            if client_max_window_bits is True:
                client_max_window_bits = self.client_max_window_bits

        else:
            if client_max_window_bits is None:
                if self.require_client_max_window_bits:
                    raise NegotiationError("required client_max_window_bits")
            elif client_max_window_bits is True:
                client_max_window_bits = self.client_max_window_bits
            elif self.client_max_window_bits < client_max_window_bits:
                client_max_window_bits = self.client_max_window_bits

        return (
            _build_parameters(
                server_no_context_takeover,
                client_no_context_takeover,
                server_max_window_bits,
                client_max_window_bits,
            ),
            PerMessageDeflate(
                client_no_context_takeover,  # remote_no_context_takeover
                server_no_context_takeover,  # local_no_context_takeover
                client_max_window_bits or 15,  # remote_max_window_bits
                server_max_window_bits or 15,  # local_max_window_bits
                self.compress_settings,
            ),
        )


def enable_server_permessage_deflate(
    extensions: Sequence[ServerExtensionFactory] | None,
) -> Sequence[ServerExtensionFactory]:
    """
    Enable Per-Message Deflate with default settings in server extensions.

    If the extension is already present, perhaps with non-default settings,
    the configuration isn't changed.

    """
    if extensions is None:
        extensions = []
    if not any(
        ext_factory.name == ServerPerMessageDeflateFactory.name
        for ext_factory in extensions
    ):
        extensions = list(extensions) + [
            ServerPerMessageDeflateFactory(
                server_max_window_bits=12,
                client_max_window_bits=12,
                compress_settings={"memLevel": 5},
            )
        ]
    return extensions



from __future__ import annotations

import asyncio
import os
import re

from ..datastructures import Headers
from ..exceptions import SecurityError


__all__ = ["read_request", "read_response"]

MAX_NUM_HEADERS = int(os.environ.get("WEBSOCKETS_MAX_NUM_HEADERS", "128"))
MAX_LINE_LENGTH = int(os.environ.get("WEBSOCKETS_MAX_LINE_LENGTH", "8192"))


def d(value: bytes) -> str:
    """
    Decode a bytestring for interpolating into an error message.

    """
    return value.decode(errors="backslashreplace")


# See https://datatracker.ietf.org/doc/html/rfc7230#appendix-B.

# Regex for validating header names.

_token_re = re.compile(rb"[-!#$%&\'*+.^_`|~0-9a-zA-Z]+")

# Regex for validating header values.

# We don't attempt to support obsolete line folding.

# Include HTAB (\x09), SP (\x20), VCHAR (\x21-\x7e), obs-text (\x80-\xff).

# The ABNF is complicated because it attempts to express that optional
# whitespace is ignored. We strip whitespace and don't revalidate that.

# See also https://www.rfc-editor.org/errata_search.php?rfc=7230&eid=4189

_value_re = re.compile(rb"[\x09\x20-\x7e\x80-\xff]*")


async def read_request(stream: asyncio.StreamReader) -> tuple[str, Headers]:
    """
    Read an HTTP/1.1 GET request and return ``(path, headers)``.

    ``path`` isn't URL-decoded or validated in any way.

    ``path`` and ``headers`` are expected to contain only ASCII characters.
    Other characters are represented with surrogate escapes.

    :func:`read_request` doesn't attempt to read the request body because
    WebSocket handshake requests don't have one. If the request contains a
    body, it may be read from ``stream`` after this coroutine returns.

    Args:
        stream: Input to read the request from.

    Raises:
        EOFError: If the connection is closed without a full HTTP request.
        SecurityError: If the request exceeds a security limit.
        ValueError: If the request isn't well formatted.

    """
    # https://datatracker.ietf.org/doc/html/rfc7230#section-3.1.1

    # Parsing is simple because fixed values are expected for method and
    # version and because path isn't checked. Since WebSocket software tends
    # to implement HTTP/1.1 strictly, there's little need for lenient parsing.

    try:
        request_line = await read_line(stream)
    except EOFError as exc:
        raise EOFError("connection closed while reading HTTP request line") from exc

    try:
        method, raw_path, version = request_line.split(b" ", 2)
    except ValueError:  # not enough values to unpack (expected 3, got 1-2)
        raise ValueError(f"invalid HTTP request line: {d(request_line)}") from None

    if method != b"GET":
        raise ValueError(f"unsupported HTTP method: {d(method)}")
    if version != b"HTTP/1.1":
        raise ValueError(f"unsupported HTTP version: {d(version)}")
    path = raw_path.decode("ascii", "surrogateescape")

    headers = await read_headers(stream)

    return path, headers


async def read_response(stream: asyncio.StreamReader) -> tuple[int, str, Headers]:
    """
    Read an HTTP/1.1 response and return ``(status_code, reason, headers)``.

    ``reason`` and ``headers`` are expected to contain only ASCII characters.
    Other characters are represented with surrogate escapes.

    :func:`read_request` doesn't attempt to read the response body because
    WebSocket handshake responses don't have one. If the response contains a
    body, it may be read from ``stream`` after this coroutine returns.

    Args:
        stream: Input to read the response from.

    Raises:
        EOFError: If the connection is closed without a full HTTP response.
        SecurityError: If the response exceeds a security limit.
        ValueError: If the response isn't well formatted.

    """
    # https://datatracker.ietf.org/doc/html/rfc7230#section-3.1.2

    # As in read_request, parsing is simple because a fixed value is expected
    # for version, status_code is a 3-digit number, and reason can be ignored.

    try:
        status_line = await read_line(stream)
    except EOFError as exc:
        raise EOFError("connection closed while reading HTTP status line") from exc

    try:
        version, raw_status_code, raw_reason = status_line.split(b" ", 2)
    except ValueError:  # not enough values to unpack (expected 3, got 1-2)
        raise ValueError(f"invalid HTTP status line: {d(status_line)}") from None

    if version != b"HTTP/1.1":
        raise ValueError(f"unsupported HTTP version: {d(version)}")
    try:
        status_code = int(raw_status_code)
    except ValueError:  # invalid literal for int() with base 10
        raise ValueError(f"invalid HTTP status code: {d(raw_status_code)}") from None
    if not 100 <= status_code < 1000:
        raise ValueError(f"unsupported HTTP status code: {d(raw_status_code)}")
    if not _value_re.fullmatch(raw_reason):
        raise ValueError(f"invalid HTTP reason phrase: {d(raw_reason)}")
    reason = raw_reason.decode()

    headers = await read_headers(stream)

    return status_code, reason, headers


async def read_headers(stream: asyncio.StreamReader) -> Headers:
    """
    Read HTTP headers from ``stream``.

    Non-ASCII characters are represented with surrogate escapes.

    """
    # https://datatracker.ietf.org/doc/html/rfc7230#section-3.2

    # We don't attempt to support obsolete line folding.

    headers = Headers()
    for _ in range(MAX_NUM_HEADERS + 1):
        try:
            line = await read_line(stream)
        except EOFError as exc:
            raise EOFError("connection closed while reading HTTP headers") from exc
        if line == b"":
            break

        try:
            raw_name, raw_value = line.split(b":", 1)
        except ValueError:  # not enough values to unpack (expected 2, got 1)
            raise ValueError(f"invalid HTTP header line: {d(line)}") from None
        if not _token_re.fullmatch(raw_name):
            raise ValueError(f"invalid HTTP header name: {d(raw_name)}")
        raw_value = raw_value.strip(b" \t")
        if not _value_re.fullmatch(raw_value):
            raise ValueError(f"invalid HTTP header value: {d(raw_value)}")

        name = raw_name.decode("ascii")  # guaranteed to be ASCII at this point
        value = raw_value.decode("ascii", "surrogateescape")
        headers[name] = value

    else:
        raise SecurityError("too many HTTP headers")

    return headers


async def read_line(stream: asyncio.StreamReader) -> bytes:
    """
    Read a single line from ``stream``.

    CRLF is stripped from the return value.

    """
    # Security: this is bounded by the StreamReader's limit (default = 32 KiB).
    line = await stream.readline()
    # Security: this guarantees header values are small (hard-coded = 8 KiB)
    if len(line) > MAX_LINE_LENGTH:
        raise SecurityError("line too long")
    # Not mandatory but safe - https://datatracker.ietf.org/doc/html/rfc7230#section-3.5
    if not line.endswith(b"\r\n"):
        raise EOFError("line without CRLF")
    return line[:-2]



from __future__ import annotations

import typing

from .imports import lazy_import
from .version import version as __version__  # noqa: F401


__all__ = [
    # .asyncio.client
    "connect",
    "unix_connect",
    # .asyncio.server
    "basic_auth",
    "broadcast",
    "serve",
    "unix_serve",
    # .client
    "ClientProtocol",
    # .datastructures
    "Headers",
    "HeadersLike",
    "MultipleValuesError",
    # .exceptions
    "ConcurrencyError",
    "ConnectionClosed",
    "ConnectionClosedError",
    "ConnectionClosedOK",
    "DuplicateParameter",
    "InvalidHandshake",
    "InvalidHeader",
    "InvalidHeaderFormat",
    "InvalidHeaderValue",
    "InvalidOrigin",
    "InvalidParameterName",
    "InvalidParameterValue",
    "InvalidState",
    "InvalidStatus",
    "InvalidUpgrade",
    "InvalidURI",
    "NegotiationError",
    "PayloadTooBig",
    "ProtocolError",
    "SecurityError",
    "WebSocketException",
    # .server
    "ServerProtocol",
    # .typing
    "Data",
    "ExtensionName",
    "ExtensionParameter",
    "LoggerLike",
    "StatusLike",
    "Origin",
    "Subprotocol",
]

# When type checking, import non-deprecated aliases eagerly. Else, import on demand.
if typing.TYPE_CHECKING:
    from .asyncio.client import connect, unix_connect
    from .asyncio.server import basic_auth, broadcast, serve, unix_serve
    from .client import ClientProtocol
    from .datastructures import Headers, HeadersLike, MultipleValuesError
    from .exceptions import (
        ConcurrencyError,
        ConnectionClosed,
        ConnectionClosedError,
        ConnectionClosedOK,
        DuplicateParameter,
        InvalidHandshake,
        InvalidHeader,
        InvalidHeaderFormat,
        InvalidHeaderValue,
        InvalidOrigin,
        InvalidParameterName,
        InvalidParameterValue,
        InvalidState,
        InvalidStatus,
        InvalidUpgrade,
        InvalidURI,
        NegotiationError,
        PayloadTooBig,
        ProtocolError,
        SecurityError,
        WebSocketException,
    )
    from .server import ServerProtocol
    from .typing import (
        Data,
        ExtensionName,
        ExtensionParameter,
        LoggerLike,
        Origin,
        StatusLike,
        Subprotocol,
    )
else:
    lazy_import(
        globals(),
        aliases={
            # .asyncio.client
            "connect": ".asyncio.client",
            "unix_connect": ".asyncio.client",
            # .asyncio.server
            "basic_auth": ".asyncio.server",
            "broadcast": ".asyncio.server",
            "serve": ".asyncio.server",
            "unix_serve": ".asyncio.server",
            # .client
            "ClientProtocol": ".client",
            # .datastructures
            "Headers": ".datastructures",
            "HeadersLike": ".datastructures",
            "MultipleValuesError": ".datastructures",
            # .exceptions
            "ConcurrencyError": ".exceptions",
            "ConnectionClosed": ".exceptions",
            "ConnectionClosedError": ".exceptions",
            "ConnectionClosedOK": ".exceptions",
            "DuplicateParameter": ".exceptions",
            "InvalidHandshake": ".exceptions",
            "InvalidHeader": ".exceptions",
            "InvalidHeaderFormat": ".exceptions",
            "InvalidHeaderValue": ".exceptions",
            "InvalidOrigin": ".exceptions",
            "InvalidParameterName": ".exceptions",
            "InvalidParameterValue": ".exceptions",
            "InvalidState": ".exceptions",
            "InvalidStatus": ".exceptions",
            "InvalidUpgrade": ".exceptions",
            "InvalidURI": ".exceptions",
            "NegotiationError": ".exceptions",
            "PayloadTooBig": ".exceptions",
            "ProtocolError": ".exceptions",
            "SecurityError": ".exceptions",
            "WebSocketException": ".exceptions",
            # .server
            "ServerProtocol": ".server",
            # .typing
            "Data": ".typing",
            "ExtensionName": ".typing",
            "ExtensionParameter": ".typing",
            "LoggerLike": ".typing",
            "Origin": ".typing",
            "StatusLike": ".typing",
            "Subprotocol": ".typing",
        },
        deprecated_aliases={
            # deprecated in 9.0 - 2021-09-01
            "framing": ".legacy",
            "handshake": ".legacy",
            "parse_uri": ".uri",
            "WebSocketURI": ".uri",
            # deprecated in 14.0 - 2024-11-09
            # .legacy.auth
            "BasicAuthWebSocketServerProtocol": ".legacy.auth",
            "basic_auth_protocol_factory": ".legacy.auth",
            # .legacy.client
            "WebSocketClientProtocol": ".legacy.client",
            # .legacy.exceptions
            "AbortHandshake": ".legacy.exceptions",
            "InvalidMessage": ".legacy.exceptions",
            "InvalidStatusCode": ".legacy.exceptions",
            "RedirectHandshake": ".legacy.exceptions",
            "WebSocketProtocolError": ".legacy.exceptions",
            # .legacy.protocol
            "WebSocketCommonProtocol": ".legacy.protocol",
            # .legacy.server
            "WebSocketServer": ".legacy.server",
            "WebSocketServerProtocol": ".legacy.server",
        },
    )



from __future__ import annotations

import argparse
import os
import signal
import sys
import threading


try:
    import readline  # noqa: F401
except ImportError:  # Windows has no `readline` normally
    pass

from .sync.client import ClientConnection, connect
from .version import version as websockets_version


if sys.platform == "win32":

    def win_enable_vt100() -> None:
        """
        Enable VT-100 for console output on Windows.

        See also https://github.com/python/cpython/issues/73245.

        """
        import ctypes

        STD_OUTPUT_HANDLE = ctypes.c_uint(-11)
        INVALID_HANDLE_VALUE = ctypes.c_uint(-1)
        ENABLE_VIRTUAL_TERMINAL_PROCESSING = 0x004

        handle = ctypes.windll.kernel32.GetStdHandle(STD_OUTPUT_HANDLE)
        if handle == INVALID_HANDLE_VALUE:
            raise RuntimeError("unable to obtain stdout handle")

        cur_mode = ctypes.c_uint()
        if ctypes.windll.kernel32.GetConsoleMode(handle, ctypes.byref(cur_mode)) == 0:
            raise RuntimeError("unable to query current console mode")

        # ctypes ints lack support for the required bit-OR operation.
        # Temporarily convert to Py int, do the OR and convert back.
        py_int_mode = int.from_bytes(cur_mode, sys.byteorder)
        new_mode = ctypes.c_uint(py_int_mode | ENABLE_VIRTUAL_TERMINAL_PROCESSING)

        if ctypes.windll.kernel32.SetConsoleMode(handle, new_mode) == 0:
            raise RuntimeError("unable to set console mode")


def print_during_input(string: str) -> None:
    sys.stdout.write(
        # Save cursor position
        "\N{ESC}7"
        # Add a new line
        "\N{LINE FEED}"
        # Move cursor up
        "\N{ESC}[A"
        # Insert blank line, scroll last line down
        "\N{ESC}[L"
        # Print string in the inserted blank line
        f"{string}\N{LINE FEED}"
        # Restore cursor position
        "\N{ESC}8"
        # Move cursor down
        "\N{ESC}[B"
    )
    sys.stdout.flush()


def print_over_input(string: str) -> None:
    sys.stdout.write(
        # Move cursor to beginning of line
        "\N{CARRIAGE RETURN}"
        # Delete current line
        "\N{ESC}[K"
        # Print string
        f"{string}\N{LINE FEED}"
    )
    sys.stdout.flush()


def print_incoming_messages(websocket: ClientConnection, stop: threading.Event) -> None:
    for message in websocket:
        if isinstance(message, str):
            print_during_input("< " + message)
        else:
            print_during_input("< (binary) " + message.hex())
    if not stop.is_set():
        # When the server closes the connection, raise KeyboardInterrupt
        # in the main thread to exit the program.
        if sys.platform == "win32":
            ctrl_c = signal.CTRL_C_EVENT
        else:
            ctrl_c = signal.SIGINT
        os.kill(os.getpid(), ctrl_c)


def main() -> None:
    # Parse command line arguments.
    parser = argparse.ArgumentParser(
        prog="python -m websockets",
        description="Interactive WebSocket client.",
        add_help=False,
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--version", action="store_true")
    group.add_argument("uri", metavar="<uri>", nargs="?")
    args = parser.parse_args()

    if args.version:
        print(f"websockets {websockets_version}")
        return

    if args.uri is None:
        parser.error("the following arguments are required: <uri>")

    # If we're on Windows, enable VT100 terminal support.
    if sys.platform == "win32":
        try:
            win_enable_vt100()
        except RuntimeError as exc:
            sys.stderr.write(
                f"Unable to set terminal to VT100 mode. This is only "
                f"supported since Win10 anniversary update. Expect "
                f"weird symbols on the terminal.\nError: {exc}\n"
            )
            sys.stderr.flush()

    try:
        websocket = connect(args.uri)
    except Exception as exc:
        print(f"Failed to connect to {args.uri}: {exc}.")
        sys.exit(1)
    else:
        print(f"Connected to {args.uri}.")

    stop = threading.Event()

    # Start the thread that reads messages from the connection.
    thread = threading.Thread(target=print_incoming_messages, args=(websocket, stop))
    thread.start()

    # Read from stdin in the main thread in order to receive signals.
    try:
        while True:
            # Since there's no size limit, put_nowait is identical to put.
            message = input("> ")
            websocket.send(message)
    except (KeyboardInterrupt, EOFError):  # ^C, ^D
        stop.set()
        websocket.close()
        print_over_input("Connection closed.")

    thread.join()


if __name__ == "__main__":
    main()



__all__ = ["PLAYER1", "PLAYER2", "Connect4"]

PLAYER1, PLAYER2 = "red", "yellow"


class Connect4:
    """
    A Connect Four game.

    Play moves with :meth:`play`.

    Get past moves with :attr:`moves`.

    Check for a victory with :attr:`winner`.

    """

    def __init__(self):
        self.moves = []
        self.top = [0 for _ in range(7)]
        self.winner = None

    @property
    def last_player(self):
        """
        Player who played the last move.

        """
        return PLAYER1 if len(self.moves) % 2 else PLAYER2

    @property
    def last_player_won(self):
        """
        Whether the last move is winning.

        """
        b = sum(1 << (8 * column + row) for _, column, row in self.moves[::-2])
        return any(b & b >> v & b >> 2 * v & b >> 3 * v for v in [1, 7, 8, 9])

    def play(self, player, column):
        """
        Play a move in a column.

        Returns the row where the checker lands.

        Raises :exc:`ValueError` if the move is illegal.

        """
        if player == self.last_player:
            raise ValueError("It isn't your turn.")

        row = self.top[column]
        if row == 6:
            raise ValueError("This slot is full.")

        self.moves.append((player, column, row))
        self.top[column] += 1

        if self.winner is None and self.last_player_won:
            self.winner = self.last_player

        return row



import contextlib
import dataclasses
import logging
import types
import unittest
from unittest.mock import patch

from websockets.client import *
from websockets.client import backoff
from websockets.datastructures import Headers
from websockets.exceptions import InvalidHandshake, InvalidHeader, InvalidStatus
from websockets.frames import OP_TEXT, Frame
from websockets.http11 import Request, Response
from websockets.protocol import CONNECTING, OPEN
from websockets.uri import parse_uri
from websockets.utils import accept_key

from .extensions.utils import (
    ClientOpExtensionFactory,
    ClientRsv2ExtensionFactory,
    OpExtension,
    Rsv2Extension,
)
from .test_utils import ACCEPT, KEY
from .utils import DATE, DeprecationTestCase


URI = parse_uri("wss://example.com/test")  # for tests where the URI doesn't matter


@patch("websockets.client.generate_key", return_value=KEY)
class BasicTests(unittest.TestCase):
    """Test basic opening handshake scenarios."""

    def test_send_request(self, _generate_key):
        """Client sends a handshake request."""
        client = ClientProtocol(URI)
        request = client.connect()
        client.send_request(request)

        self.assertEqual(
            client.data_to_send(),
            [
                f"GET /test HTTP/1.1\r\n"
                f"Host: example.com\r\n"
                f"Upgrade: websocket\r\n"
                f"Connection: Upgrade\r\n"
                f"Sec-WebSocket-Key: {KEY}\r\n"
                f"Sec-WebSocket-Version: 13\r\n"
                f"\r\n".encode()
            ],
        )
        self.assertFalse(client.close_expected())
        self.assertEqual(client.state, CONNECTING)

    def test_receive_successful_response(self, _generate_key):
        """Client receives a successful handshake response."""
        client = ClientProtocol(URI)
        client.receive_data(
            (
                f"HTTP/1.1 101 Switching Protocols\r\n"
                f"Upgrade: websocket\r\n"
                f"Connection: Upgrade\r\n"
                f"Sec-WebSocket-Accept: {ACCEPT}\r\n"
                f"Date: {DATE}\r\n"
                f"\r\n"
            ).encode(),
        )

        self.assertEqual(client.data_to_send(), [])
        self.assertFalse(client.close_expected())
        self.assertEqual(client.state, OPEN)

    def test_receive_failed_response(self, _generate_key):
        """Client receives a failed handshake response."""
        client = ClientProtocol(URI)
        client.receive_data(
            (
                f"HTTP/1.1 404 Not Found\r\n"
                f"Date: {DATE}\r\n"
                f"Content-Length: 13\r\n"
                f"Content-Type: text/plain; charset=utf-8\r\n"
                f"Connection: close\r\n"
                f"\r\n"
                f"Sorry folks.\n"
            ).encode(),
        )

        self.assertEqual(client.data_to_send(), [b""])
        self.assertTrue(client.close_expected())
        self.assertEqual(client.state, CONNECTING)


class RequestTests(unittest.TestCase):
    """Test generating opening handshake requests."""

    @patch("websockets.client.generate_key", return_value=KEY)
    def test_connect(self, _generate_key):
        """connect() creates an opening handshake request."""
        client = ClientProtocol(URI)
        request = client.connect()

        self.assertIsInstance(request, Request)
        self.assertEqual(request.path, "/test")
        self.assertEqual(
            request.headers,
            Headers(
                {
                    "Host": "example.com",
                    "Upgrade": "websocket",
                    "Connection": "Upgrade",
                    "Sec-WebSocket-Key": KEY,
                    "Sec-WebSocket-Version": "13",
                }
            ),
        )

    def test_path(self):
        """connect() uses the path from the URI."""
        client = ClientProtocol(parse_uri("wss://example.com/endpoint?test=1"))
        request = client.connect()

        self.assertEqual(request.path, "/endpoint?test=1")

    def test_port(self):
        """connect() uses the port from the URI or the default port."""
        for uri, host in [
            ("ws://example.com/", "example.com"),
            ("ws://example.com:80/", "example.com"),
            ("ws://example.com:8080/", "example.com:8080"),
            ("wss://example.com/", "example.com"),
            ("wss://example.com:443/", "example.com"),
            ("wss://example.com:8443/", "example.com:8443"),
        ]:
            with self.subTest(uri=uri):
                client = ClientProtocol(parse_uri(uri))
                request = client.connect()

                self.assertEqual(request.headers["Host"], host)

    def test_user_info(self):
        """connect() perfoms HTTP Basic Authentication with user info from the URI."""
        client = ClientProtocol(parse_uri("wss://hello:iloveyou@example.com/"))
        request = client.connect()

        self.assertEqual(request.headers["Authorization"], "Basic aGVsbG86aWxvdmV5b3U=")

    def test_origin(self):
        """connect(origin=...) generates an Origin header."""
        client = ClientProtocol(URI, origin="https://example.com")
        request = client.connect()

        self.assertEqual(request.headers["Origin"], "https://example.com")

    def test_extensions(self):
        """connect(extensions=...) generates a Sec-WebSocket-Extensions header."""
        client = ClientProtocol(URI, extensions=[ClientOpExtensionFactory()])
        request = client.connect()

        self.assertEqual(request.headers["Sec-WebSocket-Extensions"], "x-op; op")

    def test_subprotocols(self):
        """connect(subprotocols=...) generates a Sec-WebSocket-Protocol header."""
        client = ClientProtocol(URI, subprotocols=["chat"])
        request = client.connect()

        self.assertEqual(request.headers["Sec-WebSocket-Protocol"], "chat")


@patch("websockets.client.generate_key", return_value=KEY)
class ResponseTests(unittest.TestCase):
    """Test receiving opening handshake responses."""

    def test_receive_successful_response(self, _generate_key):
        """Client receives a successful handshake response."""
        client = ClientProtocol(URI)
        client.receive_data(
            (
                f"HTTP/1.1 101 Switching Protocols\r\n"
                f"Upgrade: websocket\r\n"
                f"Connection: Upgrade\r\n"
                f"Sec-WebSocket-Accept: {ACCEPT}\r\n"
                f"Date: {DATE}\r\n"
                f"\r\n"
            ).encode(),
        )
        [response] = client.events_received()

        self.assertEqual(response.status_code, 101)
        self.assertEqual(response.reason_phrase, "Switching Protocols")
        self.assertEqual(
            response.headers,
            Headers(
                {
                    "Upgrade": "websocket",
                    "Connection": "Upgrade",
                    "Sec-WebSocket-Accept": ACCEPT,
                    "Date": DATE,
                }
            ),
        )
        self.assertIsNone(response.body)
        self.assertIsNone(client.handshake_exc)

    def test_receive_failed_response(self, _generate_key):
        """Client receives a failed handshake response."""
        client = ClientProtocol(URI)
        client.receive_data(
            (
                f"HTTP/1.1 404 Not Found\r\n"
                f"Date: {DATE}\r\n"
                f"Content-Length: 13\r\n"
                f"Content-Type: text/plain; charset=utf-8\r\n"
                f"Connection: close\r\n"
                f"\r\n"
                f"Sorry folks.\n"
            ).encode(),
        )
        [response] = client.events_received()

        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.reason_phrase, "Not Found")
        self.assertEqual(
            response.headers,
            Headers(
                {
                    "Date": DATE,
                    "Content-Length": "13",
                    "Content-Type": "text/plain; charset=utf-8",
                    "Connection": "close",
                }
            ),
        )
        self.assertEqual(response.body, b"Sorry folks.\n")
        self.assertIsInstance(client.handshake_exc, InvalidStatus)
        self.assertEqual(
            str(client.handshake_exc),
            "server rejected WebSocket connection: HTTP 404",
        )

    def test_receive_no_response(self, _generate_key):
        """Client receives no handshake response."""
        client = ClientProtocol(URI)
        client.receive_eof()

        self.assertEqual(client.events_received(), [])
        self.assertIsInstance(client.handshake_exc, EOFError)
        self.assertEqual(
            str(client.handshake_exc),
            "connection closed while reading HTTP status line",
        )

    def test_receive_truncated_response(self, _generate_key):
        """Client receives a truncated handshake response."""
        client = ClientProtocol(URI)
        client.receive_data(b"HTTP/1.1 101 Switching Protocols\r\n")
        client.receive_eof()

        self.assertEqual(client.events_received(), [])
        self.assertIsInstance(client.handshake_exc, EOFError)
        self.assertEqual(
            str(client.handshake_exc),
            "connection closed while reading HTTP headers",
        )

    def test_receive_random_response(self, _generate_key):
        """Client receives a junk handshake response."""
        client = ClientProtocol(URI)
        client.receive_data(b"220 smtp.invalid\r\n")
        client.receive_data(b"250 Hello relay.invalid\r\n")
        client.receive_data(b"250 Ok\r\n")
        client.receive_data(b"250 Ok\r\n")

        self.assertEqual(client.events_received(), [])
        self.assertIsInstance(client.handshake_exc, ValueError)
        self.assertEqual(
            str(client.handshake_exc),
            "invalid HTTP status line: 220 smtp.invalid",
        )


@contextlib.contextmanager
def alter_and_receive_response(client):
    """Generate a handshake response that can be altered for testing."""
    # We could start by sending a handshake request, i.e.:
    # request = client.connect()
    # client.send_request(request)
    # However, in the current implementation, these calls have no effect on the
    # state of the client. Therefore, they're unnecessary and can be skipped.
    response = Response(
        status_code=101,
        reason_phrase="Switching Protocols",
        headers=Headers(
            {
                "Upgrade": "websocket",
                "Connection": "Upgrade",
                "Sec-WebSocket-Accept": accept_key(client.key),
            }
        ),
    )
    yield response
    client.receive_data(response.serialize())
    [parsed_response] = client.events_received()
    assert response == dataclasses.replace(parsed_response, _exception=None)


class HandshakeTests(unittest.TestCase):
    """Test processing of handshake responses to configure the connection."""

    def assertHandshakeSuccess(self, client):
        """Assert that the opening handshake succeeded."""
        self.assertEqual(client.state, OPEN)
        self.assertIsNone(client.handshake_exc)

    def assertHandshakeError(self, client, exc_type, msg):
        """Assert that the opening handshake failed with the given exception."""
        self.assertEqual(client.state, CONNECTING)
        self.assertIsInstance(client.handshake_exc, exc_type)
        # Exception chaining isn't used is client handshake implementation.
        assert client.handshake_exc.__cause__ is None
        self.assertEqual(str(client.handshake_exc), msg)

    def test_basic(self):
        """Handshake succeeds."""
        client = ClientProtocol(URI)
        with alter_and_receive_response(client):
            pass

        self.assertHandshakeSuccess(client)

    def test_missing_connection(self):
        """Handshake fails when the Connection header is missing."""
        client = ClientProtocol(URI)
        with alter_and_receive_response(client) as response:
            del response.headers["Connection"]

        self.assertHandshakeError(
            client,
            InvalidHeader,
            "missing Connection header",
        )

    def test_invalid_connection(self):
        """Handshake fails when the Connection header is invalid."""
        client = ClientProtocol(URI)
        with alter_and_receive_response(client) as response:
            del response.headers["Connection"]
            response.headers["Connection"] = "close"

        self.assertHandshakeError(
            client,
            InvalidHeader,
            "invalid Connection header: close",
        )

    def test_missing_upgrade(self):
        """Handshake fails when the Upgrade header is missing."""
        client = ClientProtocol(URI)
        with alter_and_receive_response(client) as response:
            del response.headers["Upgrade"]

        self.assertHandshakeError(
            client,
            InvalidHeader,
            "missing Upgrade header",
        )

    def test_invalid_upgrade(self):
        """Handshake fails when the Upgrade header is invalid."""
        client = ClientProtocol(URI)
        with alter_and_receive_response(client) as response:
            del response.headers["Upgrade"]
            response.headers["Upgrade"] = "h2c"

        self.assertHandshakeError(
            client,
            InvalidHeader,
            "invalid Upgrade header: h2c",
        )

    def test_missing_accept(self):
        """Handshake fails when the Sec-WebSocket-Accept header is missing."""
        client = ClientProtocol(URI)
        with alter_and_receive_response(client) as response:
            del response.headers["Sec-WebSocket-Accept"]

        self.assertHandshakeError(
            client,
            InvalidHeader,
            "missing Sec-WebSocket-Accept header",
        )

    def test_multiple_accept(self):
        """Handshake fails when the Sec-WebSocket-Accept header is repeated."""
        client = ClientProtocol(URI)
        with alter_and_receive_response(client) as response:
            response.headers["Sec-WebSocket-Accept"] = ACCEPT

        self.assertHandshakeError(
            client,
            InvalidHeader,
            "invalid Sec-WebSocket-Accept header: multiple values",
        )

    def test_invalid_accept(self):
        """Handshake fails when the Sec-WebSocket-Accept header is invalid."""
        client = ClientProtocol(URI)
        with alter_and_receive_response(client) as response:
            del response.headers["Sec-WebSocket-Accept"]
            response.headers["Sec-WebSocket-Accept"] = ACCEPT

        self.assertHandshakeError(
            client,
            InvalidHeader,
            f"invalid Sec-WebSocket-Accept header: {ACCEPT}",
        )

    def test_no_extensions(self):
        """Handshake succeeds without extensions."""
        client = ClientProtocol(URI)
        with alter_and_receive_response(client):
            pass

        self.assertHandshakeSuccess(client)
        self.assertEqual(client.extensions, [])

    def test_offer_extension(self):
        """Client offers an extension."""
        client = ClientProtocol(URI, extensions=[ClientRsv2ExtensionFactory()])
        request = client.connect()

        self.assertEqual(request.headers["Sec-WebSocket-Extensions"], "x-rsv2")

    def test_enable_extension(self):
        """Client offers an extension and the server enables it."""
        client = ClientProtocol(URI, extensions=[ClientRsv2ExtensionFactory()])
        with alter_and_receive_response(client) as response:
            response.headers["Sec-WebSocket-Extensions"] = "x-rsv2"

        self.assertHandshakeSuccess(client)
        self.assertEqual(client.extensions, [Rsv2Extension()])

    def test_extension_not_enabled(self):
        """Client offers an extension, but the server doesn't enable it."""
        client = ClientProtocol(URI, extensions=[ClientRsv2ExtensionFactory()])
        with alter_and_receive_response(client):
            pass

        self.assertHandshakeSuccess(client)
        self.assertEqual(client.extensions, [])

    def test_no_extensions_offered(self):
        """Server enables an extension when the client didn't offer any."""
        client = ClientProtocol(URI)
        with alter_and_receive_response(client) as response:
            response.headers["Sec-WebSocket-Extensions"] = "x-rsv2"

        self.assertHandshakeError(
            client,
            InvalidHandshake,
            "no extensions supported",
        )

    def test_extension_not_offered(self):
        """Server enables an extension that the client didn't offer."""
        client = ClientProtocol(URI, extensions=[ClientRsv2ExtensionFactory()])
        with alter_and_receive_response(client) as response:
            response.headers["Sec-WebSocket-Extensions"] = "x-op; op"

        self.assertHandshakeError(
            client,
            InvalidHandshake,
            "Unsupported extension: name = x-op, params = [('op', None)]",
        )

    def test_supported_extension_parameters(self):
        """Server enables an extension with parameters supported by the client."""
        client = ClientProtocol(URI, extensions=[ClientOpExtensionFactory("this")])
        with alter_and_receive_response(client) as response:
            response.headers["Sec-WebSocket-Extensions"] = "x-op; op=this"

        self.assertHandshakeSuccess(client)
        self.assertEqual(client.extensions, [OpExtension("this")])

    def test_unsupported_extension_parameters(self):
        """Server enables an extension with parameters unsupported by the client."""
        client = ClientProtocol(URI, extensions=[ClientOpExtensionFactory("this")])
        with alter_and_receive_response(client) as response:
            response.headers["Sec-WebSocket-Extensions"] = "x-op; op=that"

        self.assertHandshakeError(
            client,
            InvalidHandshake,
            "Unsupported extension: name = x-op, params = [('op', 'that')]",
        )

    def test_multiple_supported_extension_parameters(self):
        """Client offers the same extension with several parameters."""
        client = ClientProtocol(
            URI,
            extensions=[
                ClientOpExtensionFactory("this"),
                ClientOpExtensionFactory("that"),
            ],
        )
        with alter_and_receive_response(client) as response:
            response.headers["Sec-WebSocket-Extensions"] = "x-op; op=that"

        self.assertHandshakeSuccess(client)
        self.assertEqual(client.extensions, [OpExtension("that")])

    def test_multiple_extensions(self):
        """Client offers several extensions and the server enables them."""
        client = ClientProtocol(
            URI,
            extensions=[
                ClientOpExtensionFactory(),
                ClientRsv2ExtensionFactory(),
            ],
        )
        with alter_and_receive_response(client) as response:
            response.headers["Sec-WebSocket-Extensions"] = "x-op; op"
            response.headers["Sec-WebSocket-Extensions"] = "x-rsv2"

        self.assertHandshakeSuccess(client)
        self.assertEqual(client.extensions, [OpExtension(), Rsv2Extension()])

    def test_multiple_extensions_order(self):
        """Client respects the order of extensions chosen by the server."""
        client = ClientProtocol(
            URI,
            extensions=[
                ClientOpExtensionFactory(),
                ClientRsv2ExtensionFactory(),
            ],
        )
        with alter_and_receive_response(client) as response:
            response.headers["Sec-WebSocket-Extensions"] = "x-rsv2"
            response.headers["Sec-WebSocket-Extensions"] = "x-op; op"

        self.assertHandshakeSuccess(client)
        self.assertEqual(client.extensions, [Rsv2Extension(), OpExtension()])

    def test_no_subprotocols(self):
        """Handshake succeeds without subprotocols."""
        client = ClientProtocol(URI)
        with alter_and_receive_response(client):
            pass

        self.assertHandshakeSuccess(client)
        self.assertIsNone(client.subprotocol)

    def test_no_subprotocol_requested(self):
        """Client doesn't offer a subprotocol, but the server enables one."""
        client = ClientProtocol(URI)
        with alter_and_receive_response(client) as response:
            response.headers["Sec-WebSocket-Protocol"] = "chat"

        self.assertHandshakeError(
            client,
            InvalidHandshake,
            "no subprotocols supported",
        )

    def test_offer_subprotocol(self):
        """Client offers a subprotocol."""
        client = ClientProtocol(URI, subprotocols=["chat"])
        request = client.connect()

        self.assertEqual(request.headers["Sec-WebSocket-Protocol"], "chat")

    def test_enable_subprotocol(self):
        """Client offers a subprotocol and the server enables it."""
        client = ClientProtocol(URI, subprotocols=["chat"])
        with alter_and_receive_response(client) as response:
            response.headers["Sec-WebSocket-Protocol"] = "chat"

        self.assertHandshakeSuccess(client)
        self.assertEqual(client.subprotocol, "chat")

    def test_no_subprotocol_accepted(self):
        """Client offers a subprotocol, but the server doesn't enable it."""
        client = ClientProtocol(URI, subprotocols=["chat"])
        with alter_and_receive_response(client):
            pass

        self.assertHandshakeSuccess(client)
        self.assertIsNone(client.subprotocol)

    def test_multiple_subprotocols(self):
        """Client offers several subprotocols and the server enables one."""
        client = ClientProtocol(URI, subprotocols=["superchat", "chat"])
        with alter_and_receive_response(client) as response:
            response.headers["Sec-WebSocket-Protocol"] = "chat"

        self.assertHandshakeSuccess(client)
        self.assertEqual(client.subprotocol, "chat")

    def test_unsupported_subprotocol(self):
        """Client offers subprotocols but the server enables another one."""
        client = ClientProtocol(URI, subprotocols=["superchat", "chat"])
        with alter_and_receive_response(client) as response:
            response.headers["Sec-WebSocket-Protocol"] = "otherchat"

        self.assertHandshakeError(
            client,
            InvalidHandshake,
            "unsupported subprotocol: otherchat",
        )

    def test_multiple_subprotocols_accepted(self):
        """Server attempts to enable multiple subprotocols."""
        client = ClientProtocol(URI, subprotocols=["superchat", "chat"])
        with alter_and_receive_response(client) as response:
            response.headers["Sec-WebSocket-Protocol"] = "superchat"
            response.headers["Sec-WebSocket-Protocol"] = "chat"

        self.assertHandshakeError(
            client,
            InvalidHandshake,
            "invalid Sec-WebSocket-Protocol header: "
            "multiple values: superchat, chat",
        )


class MiscTests(unittest.TestCase):
    def test_bypass_handshake(self):
        """ClientProtocol bypasses the opening handshake."""
        client = ClientProtocol(URI, state=OPEN)
        client.receive_data(b"\x81\x06Hello!")
        [frame] = client.events_received()
        self.assertEqual(frame, Frame(OP_TEXT, b"Hello!"))

    def test_custom_logger(self):
        """ClientProtocol accepts a logger argument."""
        logger = logging.getLogger("test")
        with self.assertLogs("test", logging.DEBUG) as logs:
            ClientProtocol(URI, logger=logger)
        self.assertEqual(len(logs.records), 1)


class BackwardsCompatibilityTests(DeprecationTestCase):
    def test_client_connection_class(self):
        """ClientConnection is a deprecated alias for ClientProtocol."""
        with self.assertDeprecationWarning(
            "ClientConnection was renamed to ClientProtocol"
        ):
            from websockets.client import ClientConnection

            client = ClientConnection("ws://localhost/")

        self.assertIsInstance(client, ClientProtocol)


class BackoffTests(unittest.TestCase):
    def test_backoff(self):
        """backoff() yields a random delay, then exponentially increasing delays."""
        backoff_gen = backoff()
        self.assertIsInstance(backoff_gen, types.GeneratorType)

        initial_delay = next(backoff_gen)
        self.assertGreaterEqual(initial_delay, 0)
        self.assertLess(initial_delay, 5)

        following_delays = [int(next(backoff_gen)) for _ in range(9)]
        self.assertEqual(following_delays, [3, 5, 8, 13, 21, 34, 55, 89, 90])



import unittest

import websockets
import websockets.asyncio.client
import websockets.asyncio.server
import websockets.client
import websockets.datastructures
import websockets.exceptions
import websockets.server
import websockets.typing
import websockets.uri


combined_exports = (
    []
    + websockets.asyncio.client.__all__
    + websockets.asyncio.server.__all__
    + websockets.client.__all__
    + websockets.datastructures.__all__
    + websockets.exceptions.__all__
    + websockets.server.__all__
    + websockets.typing.__all__
)

# These API are intentionally not re-exported by the top-level module.
missing_reexports = [
    # websockets.asyncio.client
    "ClientConnection",
    # websockets.asyncio.server
    "ServerConnection",
    "Server",
]


class ExportsTests(unittest.TestCase):
    def test_top_level_module_reexports_submodule_exports(self):
        self.assertEqual(
            set(combined_exports),
            set(websockets.__all__ + missing_reexports),
        )

    def test_submodule_exports_are_globally_unique(self):
        self.assertEqual(
            len(set(combined_exports)),
            len(combined_exports),
        )



from __future__ import annotations

import sys


__all__ = ["TimeoutError", "aiter", "anext", "asyncio_timeout", "asyncio_timeout_at"]


if sys.version_info[:2] >= (3, 11):
    TimeoutError = TimeoutError
    aiter = aiter
    anext = anext
    from asyncio import (
        timeout as asyncio_timeout,  # noqa: F401
        timeout_at as asyncio_timeout_at,  # noqa: F401
    )

else:  # Python < 3.11
    from asyncio import TimeoutError

    def aiter(async_iterable):
        return type(async_iterable).__aiter__(async_iterable)

    async def anext(async_iterator):
        return await type(async_iterator).__anext__(async_iterator)

    from .async_timeout import (
        timeout as asyncio_timeout,  # noqa: F401
        timeout_at as asyncio_timeout_at,  # noqa: F401
    )



#!/usr/bin/env python

"""Echo server using the threading API."""

from websockets.sync.server import serve


def echo(websocket):
    for message in websocket:
        websocket.send(message)


def main():
    with serve(echo, "localhost", 8765) as server:
        server.serve_forever()


if __name__ == "__main__":
    main()



#!/usr/bin/env python

# WS server example listening on a Unix socket

import asyncio
import os.path

from websockets.legacy.server import unix_serve

async def hello(websocket):
    name = await websocket.recv()
    print(f"<<< {name}")

    greeting = f"Hello {name}!"

    await websocket.send(greeting)
    print(f">>> {greeting}")

async def main():
    socket_path = os.path.join(os.path.dirname(__file__), "socket")
    async with unix_serve(hello, socket_path):
        await asyncio.get_running_loop().create_future()  # run forever

asyncio.run(main())



#!/usr/bin/env python

import asyncio
import json
import secrets

from websockets.asyncio.server import broadcast, serve

from connect4 import PLAYER1, PLAYER2, Connect4


JOIN = {}

WATCH = {}


async def error(websocket, message):
    """
    Send an error message.

    """
    event = {
        "type": "error",
        "message": message,
    }
    await websocket.send(json.dumps(event))


async def replay(websocket, game):
    """
    Send previous moves.

    """
    # Make a copy to avoid an exception if game.moves changes while iteration
    # is in progress. If a move is played while replay is running, moves will
    # be sent out of order but each move will be sent once and eventually the
    # UI will be consistent.
    for player, column, row in game.moves.copy():
        event = {
            "type": "play",
            "player": player,
            "column": column,
            "row": row,
        }
        await websocket.send(json.dumps(event))


async def play(websocket, game, player, connected):
    """
    Receive and process moves from a player.

    """
    async for message in websocket:
        # Parse a "play" event from the UI.
        event = json.loads(message)
        assert event["type"] == "play"
        column = event["column"]

        try:
            # Play the move.
            row = game.play(player, column)
        except ValueError as exc:
            # Send an "error" event if the move was illegal.
            await error(websocket, str(exc))
            continue

        # Send a "play" event to update the UI.
        event = {
            "type": "play",
            "player": player,
            "column": column,
            "row": row,
        }
        broadcast(connected, json.dumps(event))

        # If move is winning, send a "win" event.
        if game.winner is not None:
            event = {
                "type": "win",
                "player": game.winner,
            }
            broadcast(connected, json.dumps(event))


async def start(websocket):
    """
    Handle a connection from the first player: start a new game.

    """
    # Initialize a Connect Four game, the set of WebSocket connections
    # receiving moves from this game, and secret access tokens.
    game = Connect4()
    connected = {websocket}

    join_key = secrets.token_urlsafe(12)
    JOIN[join_key] = game, connected

    watch_key = secrets.token_urlsafe(12)
    WATCH[watch_key] = game, connected

    try:
        # Send the secret access tokens to the browser of the first player,
        # where they'll be used for building "join" and "watch" links.
        event = {
            "type": "init",
            "join": join_key,
            "watch": watch_key,
        }
        await websocket.send(json.dumps(event))
        # Receive and process moves from the first player.
        await play(websocket, game, PLAYER1, connected)
    finally:
        del JOIN[join_key]
        del WATCH[watch_key]


async def join(websocket, join_key):
    """
    Handle a connection from the second player: join an existing game.

    """
    # Find the Connect Four game.
    try:
        game, connected = JOIN[join_key]
    except KeyError:
        await error(websocket, "Game not found.")
        return

    # Register to receive moves from this game.
    connected.add(websocket)
    try:
        # Send the first move, in case the first player already played it.
        await replay(websocket, game)
        # Receive and process moves from the second player.
        await play(websocket, game, PLAYER2, connected)
    finally:
        connected.remove(websocket)


async def watch(websocket, watch_key):
    """
    Handle a connection from a spectator: watch an existing game.

    """
    # Find the Connect Four game.
    try:
        game, connected = WATCH[watch_key]
    except KeyError:
        await error(websocket, "Game not found.")
        return

    # Register to receive moves from this game.
    connected.add(websocket)
    try:
        # Send previous moves, in case the game already started.
        await replay(websocket, game)
        # Keep the connection open, but don't receive any messages.
        await websocket.wait_closed()
    finally:
        connected.remove(websocket)


async def handler(websocket):
    """
    Handle a connection and dispatch it according to who is connecting.

    """
    # Receive and parse the "init" event from the UI.
    message = await websocket.recv()
    event = json.loads(message)
    assert event["type"] == "init"

    if "join" in event:
        # Second player joins an existing game.
        await join(websocket, event["join"])
    elif "watch" in event:
        # Spectator watches an existing game.
        await watch(websocket, event["watch"])
    else:
        # First player starts a new game.
        await start(websocket)


async def main():
    async with serve(handler, "", 8001):
        await asyncio.get_running_loop().create_future()  # run forever


if __name__ == "__main__":
    asyncio.run(main())



#!/usr/bin/env python

import asyncio

from websockets.asyncio.client import connect

async def hello():
    uri = "ws://localhost:8765"
    async with connect(uri) as websocket:
        name = input("What's your name? ")

        await websocket.send(name)
        print(f">>> {name}")

        greeting = await websocket.recv()
        print(f"<<< {greeting}")

if __name__ == "__main__":
    asyncio.run(hello())



#!/usr/bin/env python

"""Measure coverage of each module by its test module."""

import glob
import os.path
import subprocess
import sys


UNMAPPED_SRC_FILES = [
    "websockets/typing.py",
    "websockets/version.py",
]

UNMAPPED_TEST_FILES = [
    "tests/test_exports.py",
]


def check_environment():
    """Check that prerequisites for running this script are met."""
    try:
        import websockets  # noqa: F401
    except ImportError:
        print("failed to import websockets; is src on PYTHONPATH?")
        return False
    try:
        import coverage  # noqa: F401
    except ImportError:
        print("failed to locate Coverage.py; is it installed?")
        return False
    return True


def get_mapping(src_dir="src"):
    """Return a dict mapping each source file to its test file."""

    # List source and test files.

    src_files = glob.glob(
        os.path.join(src_dir, "websockets/**/*.py"),
        recursive=True,
    )
    test_files = glob.glob(
        "tests/**/*.py",
        recursive=True,
    )

    src_files = [
        os.path.relpath(src_file, src_dir)
        for src_file in sorted(src_files)
        if "legacy" not in os.path.dirname(src_file)
        and os.path.basename(src_file) != "__init__.py"
        and os.path.basename(src_file) != "__main__.py"
        and os.path.basename(src_file) != "async_timeout.py"
        and os.path.basename(src_file) != "compatibility.py"
    ]
    test_files = [
        test_file
        for test_file in sorted(test_files)
        if "legacy" not in os.path.dirname(test_file)
        and os.path.basename(test_file) != "__init__.py"
        and os.path.basename(test_file).startswith("test_")
    ]

    # Map source files to test files.

    mapping = {}
    unmapped_test_files = set()

    for test_file in test_files:
        dir_name, file_name = os.path.split(test_file)
        assert dir_name.startswith("tests")
        assert file_name.startswith("test_")
        src_file = os.path.join(
            "websockets" + dir_name[len("tests") :],
            file_name[len("test_") :],
        )
        if src_file in src_files:
            mapping[src_file] = test_file
        else:
            unmapped_test_files.add(test_file)

    unmapped_src_files = set(src_files) - set(mapping)

    # Ensure that all files are mapped.

    assert unmapped_src_files == set(UNMAPPED_SRC_FILES)
    assert unmapped_test_files == set(UNMAPPED_TEST_FILES)

    return mapping


def get_ignored_files(src_dir="src"):
    """Return the list of files to exclude from coverage measurement."""
    # */websockets matches src/websockets and .tox/**/site-packages/websockets.
    return [
        # There are no tests for the __main__ module.
        "*/websockets/__main__.py",
        # There is nothing to test on type declarations.
        "*/websockets/typing.py",
        # We don't test compatibility modules with previous versions of Python
        # or websockets (import locations).
        "*/websockets/asyncio/async_timeout.py",
        "*/websockets/asyncio/compatibility.py",
        # This approach isn't applicable to the test suite of the legacy
        # implementation, due to the huge test_client_server test module.
        "*/websockets/legacy/*",
        "tests/legacy/*",
    ] + [
        # Exclude test utilities that are shared between several test modules.
        # Also excludes this script.
        test_file
        for test_file in sorted(glob.glob("tests/**/*.py", recursive=True))
        if "legacy" not in os.path.dirname(test_file)
        and os.path.basename(test_file) != "__init__.py"
        and not os.path.basename(test_file).startswith("test_")
    ]


def run_coverage(mapping, src_dir="src"):
    # Initialize a new coverage measurement session. The --source option
    # includes all files in the report, even if they're never imported.
    print("\nInitializing session\n", flush=True)
    subprocess.run(
        [
            sys.executable,
            "-m",
            "coverage",
            "run",
            "--source",
            ",".join([os.path.join(src_dir, "websockets"), "tests"]),
            "--omit",
            ",".join(get_ignored_files(src_dir)),
            "-m",
            "unittest",
        ]
        + list(UNMAPPED_TEST_FILES),
        check=True,
    )
    # Append coverage of each source module by the corresponding test module.
    for src_file, test_file in mapping.items():
        print(f"\nTesting {src_file} with {test_file}\n", flush=True)
        subprocess.run(
            [
                sys.executable,
                "-m",
                "coverage",
                "run",
                "--append",
                "--include",
                ",".join([os.path.join(src_dir, src_file), test_file]),
                "-m",
                "unittest",
                test_file,
            ],
            check=True,
        )


if __name__ == "__main__":
    if not check_environment():
        sys.exit(1)
    src_dir = sys.argv[1] if len(sys.argv) == 2 else "src"
    mapping = get_mapping(src_dir)
    run_coverage(mapping, src_dir)



import datetime
import json
import logging

class JSONFormatter(logging.Formatter):
    """
    Render logs as JSON.

    To add details to a log record, store them in a ``event_data``
    custom attribute. This dict is merged into the event.

    """
    def __init__(self):
        pass  # override logging.Formatter constructor

    def format(self, record):
        event = {
            "timestamp": self.getTimestamp(record.created),
            "message": record.getMessage(),
            "level": record.levelname,
            "logger": record.name,
        }
        event_data = getattr(record, "event_data", None)
        if event_data:
            event.update(event_data)
        if record.exc_info:
            event["exc_info"] = self.formatException(record.exc_info)
        if record.stack_info:
            event["stack_info"] = self.formatStack(record.stack_info)
        return json.dumps(event)

    def getTimestamp(self, created):
        return datetime.datetime.utcfromtimestamp(created).isoformat()



#!/usr/bin/env python

import asyncio
import datetime
import random

from websockets.asyncio.server import broadcast, serve

CONNECTIONS = set()

async def register(websocket):
    CONNECTIONS.add(websocket)
    try:
        await websocket.wait_closed()
    finally:
        CONNECTIONS.remove(websocket)

async def show_time():
    while True:
        message = datetime.datetime.utcnow().isoformat() + "Z"
        broadcast(CONNECTIONS, message)
        await asyncio.sleep(random.random() * 2 + 1)

async def main():
    async with serve(register, "localhost", 5678):
        await show_time()

if __name__ == "__main__":
    asyncio.run(main())



import json
import logging

from websockets.exceptions import WebSocketException
from websockets.sync.client import connect


logging.basicConfig(level=logging.WARNING)

SERVER = "ws://localhost:9001"

AGENT = "websockets.sync"


def get_case_count():
    with connect(f"{SERVER}/getCaseCount") as ws:
        return json.loads(ws.recv())


def run_case(case):
    with connect(
        f"{SERVER}/runCase?case={case}&agent={AGENT}",
        max_size=2**25,
    ) as ws:
        try:
            for msg in ws:
                ws.send(msg)
        except WebSocketException:
            pass


def update_reports():
    with connect(
        f"{SERVER}/updateReports?agent={AGENT}",
        open_timeout=60,
    ):
        pass


def main():
    cases = get_case_count()
    for case in range(1, cases + 1):
        print(f"Running test case {case:03d} / {cases}... ", end="\t")
        try:
            run_case(case)
        except WebSocketException as exc:
            print(f"ERROR: {type(exc).__name__}: {exc}")
        except Exception as exc:
            print(f"FAIL: {type(exc).__name__}: {exc}")
        else:
            print("OK")
    print(f"Ran {cases} test cases")
    update_reports()
    print("Updated reports")


if __name__ == "__main__":
    main()



import dataclasses
import os
import unittest

from websockets.exceptions import (
    DuplicateParameter,
    InvalidParameterName,
    InvalidParameterValue,
    NegotiationError,
    PayloadTooBig,
    ProtocolError,
)
from websockets.extensions.permessage_deflate import *
from websockets.frames import (
    OP_BINARY,
    OP_CLOSE,
    OP_CONT,
    OP_PING,
    OP_PONG,
    OP_TEXT,
    Close,
    CloseCode,
    Frame,
)

from .utils import ClientNoOpExtensionFactory, ServerNoOpExtensionFactory


class PerMessageDeflateTestsMixin:
    def assertExtensionEqual(self, extension1, extension2):
        self.assertEqual(
            extension1.remote_no_context_takeover,
            extension2.remote_no_context_takeover,
        )
        self.assertEqual(
            extension1.local_no_context_takeover,
            extension2.local_no_context_takeover,
        )
        self.assertEqual(
            extension1.remote_max_window_bits,
            extension2.remote_max_window_bits,
        )
        self.assertEqual(
            extension1.local_max_window_bits,
            extension2.local_max_window_bits,
        )


class PerMessageDeflateTests(unittest.TestCase, PerMessageDeflateTestsMixin):
    def setUp(self):
        # Set up an instance of the permessage-deflate extension with the most
        # common settings. Since the extension is symmetrical, this instance
        # may be used for testing both encoding and decoding.
        self.extension = PerMessageDeflate(False, False, 15, 15)

    def test_name(self):
        assert self.extension.name == "permessage-deflate"

    def test_repr(self):
        self.assertExtensionEqual(eval(repr(self.extension)), self.extension)

    # Control frames aren't encoded or decoded.

    def test_no_encode_decode_ping_frame(self):
        frame = Frame(OP_PING, b"")

        self.assertEqual(self.extension.encode(frame), frame)

        self.assertEqual(self.extension.decode(frame), frame)

    def test_no_encode_decode_pong_frame(self):
        frame = Frame(OP_PONG, b"")

        self.assertEqual(self.extension.encode(frame), frame)

        self.assertEqual(self.extension.decode(frame), frame)

    def test_no_encode_decode_close_frame(self):
        frame = Frame(OP_CLOSE, Close(CloseCode.NORMAL_CLOSURE, "").serialize())

        self.assertEqual(self.extension.encode(frame), frame)

        self.assertEqual(self.extension.decode(frame), frame)

    # Data frames are encoded and decoded.

    def test_encode_decode_text_frame(self):
        frame = Frame(OP_TEXT, "café".encode())

        enc_frame = self.extension.encode(frame)

        self.assertEqual(
            enc_frame,
            dataclasses.replace(frame, rsv1=True, data=b"JNL;\xbc\x12\x00"),
        )

        dec_frame = self.extension.decode(enc_frame)

        self.assertEqual(dec_frame, frame)

    def test_encode_decode_binary_frame(self):
        frame = Frame(OP_BINARY, b"tea")

        enc_frame = self.extension.encode(frame)

        self.assertEqual(
            enc_frame,
            dataclasses.replace(frame, rsv1=True, data=b"*IM\x04\x00"),
        )

        dec_frame = self.extension.decode(enc_frame)

        self.assertEqual(dec_frame, frame)

    def test_encode_decode_fragmented_text_frame(self):
        frame1 = Frame(OP_TEXT, "café".encode(), fin=False)
        frame2 = Frame(OP_CONT, " & ".encode(), fin=False)
        frame3 = Frame(OP_CONT, "croissants".encode())

        enc_frame1 = self.extension.encode(frame1)
        enc_frame2 = self.extension.encode(frame2)
        enc_frame3 = self.extension.encode(frame3)

        self.assertEqual(
            enc_frame1,
            dataclasses.replace(
                frame1, rsv1=True, data=b"JNL;\xbc\x12\x00\x00\x00\xff\xff"
            ),
        )
        self.assertEqual(
            enc_frame2,
            dataclasses.replace(frame2, data=b"RPS\x00\x00\x00\x00\xff\xff"),
        )
        self.assertEqual(
            enc_frame3,
            dataclasses.replace(frame3, data=b"J.\xca\xcf,.N\xcc+)\x06\x00"),
        )

        dec_frame1 = self.extension.decode(enc_frame1)
        dec_frame2 = self.extension.decode(enc_frame2)
        dec_frame3 = self.extension.decode(enc_frame3)

        self.assertEqual(dec_frame1, frame1)
        self.assertEqual(dec_frame2, frame2)
        self.assertEqual(dec_frame3, frame3)

    def test_encode_decode_fragmented_binary_frame(self):
        frame1 = Frame(OP_TEXT, b"tea ", fin=False)
        frame2 = Frame(OP_CONT, b"time")

        enc_frame1 = self.extension.encode(frame1)
        enc_frame2 = self.extension.encode(frame2)

        self.assertEqual(
            enc_frame1,
            dataclasses.replace(
                frame1, rsv1=True, data=b"*IMT\x00\x00\x00\x00\xff\xff"
            ),
        )
        self.assertEqual(
            enc_frame2,
            dataclasses.replace(frame2, data=b"*\xc9\xccM\x05\x00"),
        )

        dec_frame1 = self.extension.decode(enc_frame1)
        dec_frame2 = self.extension.decode(enc_frame2)

        self.assertEqual(dec_frame1, frame1)
        self.assertEqual(dec_frame2, frame2)

    def test_encode_decode_large_frame(self):
        # There is a separate code path that avoids copying data
        # when frames are larger than 2kB. Test it for coverage.
        frame = Frame(OP_BINARY, os.urandom(4096))

        enc_frame = self.extension.encode(frame)
        dec_frame = self.extension.decode(enc_frame)

        self.assertEqual(dec_frame, frame)

    def test_no_decode_text_frame(self):
        frame = Frame(OP_TEXT, "café".encode())

        # Try decoding a frame that wasn't encoded.
        self.assertEqual(self.extension.decode(frame), frame)

    def test_no_decode_binary_frame(self):
        frame = Frame(OP_TEXT, b"tea")

        # Try decoding a frame that wasn't encoded.
        self.assertEqual(self.extension.decode(frame), frame)

    def test_no_decode_fragmented_text_frame(self):
        frame1 = Frame(OP_TEXT, "café".encode(), fin=False)
        frame2 = Frame(OP_CONT, " & ".encode(), fin=False)
        frame3 = Frame(OP_CONT, "croissants".encode())

        dec_frame1 = self.extension.decode(frame1)
        dec_frame2 = self.extension.decode(frame2)
        dec_frame3 = self.extension.decode(frame3)

        self.assertEqual(dec_frame1, frame1)
        self.assertEqual(dec_frame2, frame2)
        self.assertEqual(dec_frame3, frame3)

    def test_no_decode_fragmented_binary_frame(self):
        frame1 = Frame(OP_TEXT, b"tea ", fin=False)
        frame2 = Frame(OP_CONT, b"time")

        dec_frame1 = self.extension.decode(frame1)
        dec_frame2 = self.extension.decode(frame2)

        self.assertEqual(dec_frame1, frame1)
        self.assertEqual(dec_frame2, frame2)

    def test_context_takeover(self):
        frame = Frame(OP_TEXT, "café".encode())

        enc_frame1 = self.extension.encode(frame)
        enc_frame2 = self.extension.encode(frame)

        self.assertEqual(enc_frame1.data, b"JNL;\xbc\x12\x00")
        self.assertEqual(enc_frame2.data, b"J\x06\x11\x00\x00")

    def test_remote_no_context_takeover(self):
        # No context takeover when decoding messages.
        self.extension = PerMessageDeflate(True, False, 15, 15)

        frame = Frame(OP_TEXT, "café".encode())

        enc_frame1 = self.extension.encode(frame)
        enc_frame2 = self.extension.encode(frame)

        self.assertEqual(enc_frame1.data, b"JNL;\xbc\x12\x00")
        self.assertEqual(enc_frame2.data, b"J\x06\x11\x00\x00")

        dec_frame1 = self.extension.decode(enc_frame1)
        self.assertEqual(dec_frame1, frame)

        with self.assertRaises(ProtocolError):
            self.extension.decode(enc_frame2)

    def test_local_no_context_takeover(self):
        # No context takeover when encoding and decoding messages.
        self.extension = PerMessageDeflate(True, True, 15, 15)

        frame = Frame(OP_TEXT, "café".encode())

        enc_frame1 = self.extension.encode(frame)
        enc_frame2 = self.extension.encode(frame)

        self.assertEqual(enc_frame1.data, b"JNL;\xbc\x12\x00")
        self.assertEqual(enc_frame2.data, b"JNL;\xbc\x12\x00")

        dec_frame1 = self.extension.decode(enc_frame1)
        dec_frame2 = self.extension.decode(enc_frame2)

        self.assertEqual(dec_frame1, frame)
        self.assertEqual(dec_frame2, frame)

    # Compression settings can be customized.

    def test_compress_settings(self):
        # Configure an extension so that no compression actually occurs.
        extension = PerMessageDeflate(False, False, 15, 15, {"level": 0})

        frame = Frame(OP_TEXT, "café".encode())

        enc_frame = extension.encode(frame)

        self.assertEqual(
            enc_frame,
            dataclasses.replace(
                frame,
                rsv1=True,
                data=b"\x00\x05\x00\xfa\xffcaf\xc3\xa9\x00",  # not compressed
            ),
        )

    # Frames aren't decoded beyond max_size.

    def test_decompress_max_size(self):
        frame = Frame(OP_TEXT, ("a" * 20).encode())

        enc_frame = self.extension.encode(frame)

        self.assertEqual(enc_frame.data, b"JL\xc4\x04\x00\x00")

        with self.assertRaises(PayloadTooBig):
            self.extension.decode(enc_frame, max_size=10)


class ClientPerMessageDeflateFactoryTests(
    unittest.TestCase, PerMessageDeflateTestsMixin
):
    def test_name(self):
        assert ClientPerMessageDeflateFactory.name == "permessage-deflate"

    def test_init(self):
        for config in [
            (False, False, 8, None),  # server_max_window_bits ≥ 8
            (False, True, 15, None),  # server_max_window_bits ≤ 15
            (True, False, None, 8),  # client_max_window_bits ≥ 8
            (True, True, None, 15),  # client_max_window_bits ≤ 15
            (False, False, None, True),  # client_max_window_bits
            (False, False, None, None, {"memLevel": 4}),
        ]:
            with self.subTest(config=config):
                # This does not raise an exception.
                ClientPerMessageDeflateFactory(*config)

    def test_init_error(self):
        for config in [
            (False, False, 7, 8),  # server_max_window_bits < 8
            (False, True, 8, 7),  # client_max_window_bits < 8
            (True, False, 16, 15),  # server_max_window_bits > 15
            (True, True, 15, 16),  # client_max_window_bits > 15
            (False, False, True, None),  # server_max_window_bits
            (False, False, None, None, {"wbits": 11}),
        ]:
            with self.subTest(config=config):
                with self.assertRaises(ValueError):
                    ClientPerMessageDeflateFactory(*config)

    def test_get_request_params(self):
        for config, result in [
            # Test without any parameter
            (
                (False, False, None, None),
                [],
            ),
            # Test server_no_context_takeover
            (
                (True, False, None, None),
                [("server_no_context_takeover", None)],
            ),
            # Test client_no_context_takeover
            (
                (False, True, None, None),
                [("client_no_context_takeover", None)],
            ),
            # Test server_max_window_bits
            (
                (False, False, 10, None),
                [("server_max_window_bits", "10")],
            ),
            # Test client_max_window_bits
            (
                (False, False, None, 10),
                [("client_max_window_bits", "10")],
            ),
            (
                (False, False, None, True),
                [("client_max_window_bits", None)],
            ),
            # Test all parameters together
            (
                (True, True, 12, 12),
                [
                    ("server_no_context_takeover", None),
                    ("client_no_context_takeover", None),
                    ("server_max_window_bits", "12"),
                    ("client_max_window_bits", "12"),
                ],
            ),
        ]:
            with self.subTest(config=config):
                factory = ClientPerMessageDeflateFactory(*config)
                self.assertEqual(factory.get_request_params(), result)

    def test_process_response_params(self):
        for config, response_params, result in [
            # Test without any parameter
            (
                (False, False, None, None),
                [],
                (False, False, 15, 15),
            ),
            (
                (False, False, None, None),
                [("unknown", None)],
                InvalidParameterName,
            ),
            # Test server_no_context_takeover
            (
                (False, False, None, None),
                [("server_no_context_takeover", None)],
                (True, False, 15, 15),
            ),
            (
                (True, False, None, None),
                [],
                NegotiationError,
            ),
            (
                (True, False, None, None),
                [("server_no_context_takeover", None)],
                (True, False, 15, 15),
            ),
            (
                (True, False, None, None),
                [("server_no_context_takeover", None)] * 2,
                DuplicateParameter,
            ),
            (
                (True, False, None, None),
                [("server_no_context_takeover", "42")],
                InvalidParameterValue,
            ),
            # Test client_no_context_takeover
            (
                (False, False, None, None),
                [("client_no_context_takeover", None)],
                (False, True, 15, 15),
            ),
            (
                (False, True, None, None),
                [],
                (False, True, 15, 15),
            ),
            (
                (False, True, None, None),
                [("client_no_context_takeover", None)],
                (False, True, 15, 15),
            ),
            (
                (False, True, None, None),
                [("client_no_context_takeover", None)] * 2,
                DuplicateParameter,
            ),
            (
                (False, True, None, None),
                [("client_no_context_takeover", "42")],
                InvalidParameterValue,
            ),
            # Test server_max_window_bits
            (
                (False, False, None, None),
                [("server_max_window_bits", "7")],
                NegotiationError,
            ),
            (
                (False, False, None, None),
                [("server_max_window_bits", "10")],
                (False, False, 10, 15),
            ),
            (
                (False, False, None, None),
                [("server_max_window_bits", "16")],
                NegotiationError,
            ),
            (
                (False, False, 12, None),
                [],
                NegotiationError,
            ),
            (
                (False, False, 12, None),
                [("server_max_window_bits", "10")],
                (False, False, 10, 15),
            ),
            (
                (False, False, 12, None),
                [("server_max_window_bits", "12")],
                (False, False, 12, 15),
            ),
            (
                (False, False, 12, None),
                [("server_max_window_bits", "13")],
                NegotiationError,
            ),
            (
                (False, False, 12, None),
                [("server_max_window_bits", "12")] * 2,
                DuplicateParameter,
            ),
            (
                (False, False, 12, None),
                [("server_max_window_bits", "42")],
                InvalidParameterValue,
            ),
            # Test client_max_window_bits
            (
                (False, False, None, None),
                [("client_max_window_bits", "10")],
                NegotiationError,
            ),
            (
                (False, False, None, True),
                [],
                (False, False, 15, 15),
            ),
            (
                (False, False, None, True),
                [("client_max_window_bits", "7")],
                NegotiationError,
            ),
            (
                (False, False, None, True),
                [("client_max_window_bits", "10")],
                (False, False, 15, 10),
            ),
            (
                (False, False, None, True),
                [("client_max_window_bits", "16")],
                NegotiationError,
            ),
            (
                (False, False, None, 12),
                [],
                (False, False, 15, 12),
            ),
            (
                (False, False, None, 12),
                [("client_max_window_bits", "10")],
                (False, False, 15, 10),
            ),
            (
                (False, False, None, 12),
                [("client_max_window_bits", "12")],
                (False, False, 15, 12),
            ),
            (
                (False, False, None, 12),
                [("client_max_window_bits", "13")],
                NegotiationError,
            ),
            (
                (False, False, None, 12),
                [("client_max_window_bits", "12")] * 2,
                DuplicateParameter,
            ),
            (
                (False, False, None, 12),
                [("client_max_window_bits", "42")],
                InvalidParameterValue,
            ),
            # Test all parameters together
            (
                (True, True, 12, 12),
                [
                    ("server_no_context_takeover", None),
                    ("client_no_context_takeover", None),
                    ("server_max_window_bits", "10"),
                    ("client_max_window_bits", "10"),
                ],
                (True, True, 10, 10),
            ),
            (
                (False, False, None, True),
                [
                    ("server_no_context_takeover", None),
                    ("client_no_context_takeover", None),
                    ("server_max_window_bits", "10"),
                    ("client_max_window_bits", "10"),
                ],
                (True, True, 10, 10),
            ),
            (
                (True, True, 12, 12),
                [
                    ("server_no_context_takeover", None),
                    ("server_max_window_bits", "12"),
                ],
                (True, True, 12, 12),
            ),
        ]:
            with self.subTest(config=config, response_params=response_params):
                factory = ClientPerMessageDeflateFactory(*config)
                if isinstance(result, type) and issubclass(result, Exception):
                    with self.assertRaises(result):
                        factory.process_response_params(response_params, [])
                else:
                    extension = factory.process_response_params(response_params, [])
                    expected = PerMessageDeflate(*result)
                    self.assertExtensionEqual(extension, expected)

    def test_process_response_params_deduplication(self):
        factory = ClientPerMessageDeflateFactory(False, False, None, None)
        with self.assertRaises(NegotiationError):
            factory.process_response_params(
                [], [PerMessageDeflate(False, False, 15, 15)]
            )

    def test_enable_client_permessage_deflate(self):
        for extensions, (
            expected_len,
            expected_position,
            expected_compress_settings,
        ) in [
            (
                None,
                (1, 0, {"memLevel": 5}),
            ),
            (
                [],
                (1, 0, {"memLevel": 5}),
            ),
            (
                [ClientNoOpExtensionFactory()],
                (2, 1, {"memLevel": 5}),
            ),
            (
                [ClientPerMessageDeflateFactory(compress_settings={"memLevel": 7})],
                (1, 0, {"memLevel": 7}),
            ),
            (
                [
                    ClientPerMessageDeflateFactory(compress_settings={"memLevel": 7}),
                    ClientNoOpExtensionFactory(),
                ],
                (2, 0, {"memLevel": 7}),
            ),
            (
                [
                    ClientNoOpExtensionFactory(),
                    ClientPerMessageDeflateFactory(compress_settings={"memLevel": 7}),
                ],
                (2, 1, {"memLevel": 7}),
            ),
        ]:
            with self.subTest(extensions=extensions):
                extensions = enable_client_permessage_deflate(extensions)
                self.assertEqual(len(extensions), expected_len)
                extension = extensions[expected_position]
                self.assertIsInstance(extension, ClientPerMessageDeflateFactory)
                self.assertEqual(
                    extension.compress_settings,
                    expected_compress_settings,
                )


class ServerPerMessageDeflateFactoryTests(
    unittest.TestCase, PerMessageDeflateTestsMixin
):
    def test_name(self):
        assert ServerPerMessageDeflateFactory.name == "permessage-deflate"

    def test_init(self):
        for config in [
            (False, False, 8, None),  # server_max_window_bits ≥ 8
            (False, True, 15, None),  # server_max_window_bits ≤ 15
            (True, False, None, 8),  # client_max_window_bits ≥ 8
            (True, True, None, 15),  # client_max_window_bits ≤ 15
            (False, False, None, None, {"memLevel": 4}),
            (False, False, None, 12, {}, True),  # require_client_max_window_bits
        ]:
            with self.subTest(config=config):
                # This does not raise an exception.
                ServerPerMessageDeflateFactory(*config)

    def test_init_error(self):
        for config in [
            (False, False, 7, 8),  # server_max_window_bits < 8
            (False, True, 8, 7),  # client_max_window_bits < 8
            (True, False, 16, 15),  # server_max_window_bits > 15
            (True, True, 15, 16),  # client_max_window_bits > 15
            (False, False, None, True),  # client_max_window_bits
            (False, False, True, None),  # server_max_window_bits
            (False, False, None, None, {"wbits": 11}),
            (False, False, None, None, {}, True),  # require_client_max_window_bits
        ]:
            with self.subTest(config=config):
                with self.assertRaises(ValueError):
                    ServerPerMessageDeflateFactory(*config)

    def test_process_request_params(self):
        # Parameters in result appear swapped vs. config because the order is
        # (remote, local) vs. (server, client).
        for config, request_params, response_params, result in [
            # Test without any parameter
            (
                (False, False, None, None),
                [],
                [],
                (False, False, 15, 15),
            ),
            (
                (False, False, None, None),
                [("unknown", None)],
                None,
                InvalidParameterName,
            ),
            # Test server_no_context_takeover
            (
                (False, False, None, None),
                [("server_no_context_takeover", None)],
                [("server_no_context_takeover", None)],
                (False, True, 15, 15),
            ),
            (
                (True, False, None, None),
                [],
                [("server_no_context_takeover", None)],
                (False, True, 15, 15),
            ),
            (
                (True, False, None, None),
                [("server_no_context_takeover", None)],
                [("server_no_context_takeover", None)],
                (False, True, 15, 15),
            ),
            (
                (True, False, None, None),
                [("server_no_context_takeover", None)] * 2,
                None,
                DuplicateParameter,
            ),
            (
                (True, False, None, None),
                [("server_no_context_takeover", "42")],
                None,
                InvalidParameterValue,
            ),
            # Test client_no_context_takeover
            (
                (False, False, None, None),
                [("client_no_context_takeover", None)],
                [("client_no_context_takeover", None)],  # doesn't matter
                (True, False, 15, 15),
            ),
            (
                (False, True, None, None),
                [],
                [("client_no_context_takeover", None)],
                (True, False, 15, 15),
            ),
            (
                (False, True, None, None),
                [("client_no_context_takeover", None)],
                [("client_no_context_takeover", None)],  # doesn't matter
                (True, False, 15, 15),
            ),
            (
                (False, True, None, None),
                [("client_no_context_takeover", None)] * 2,
                None,
                DuplicateParameter,
            ),
            (
                (False, True, None, None),
                [("client_no_context_takeover", "42")],
                None,
                InvalidParameterValue,
            ),
            # Test server_max_window_bits
            (
                (False, False, None, None),
                [("server_max_window_bits", "7")],
                None,
                NegotiationError,
            ),
            (
                (False, False, None, None),
                [("server_max_window_bits", "10")],
                [("server_max_window_bits", "10")],
                (False, False, 15, 10),
            ),
            (
                (False, False, None, None),
                [("server_max_window_bits", "16")],
                None,
                NegotiationError,
            ),
            (
                (False, False, 12, None),
                [],
                [("server_max_window_bits", "12")],
                (False, False, 15, 12),
            ),
            (
                (False, False, 12, None),
                [("server_max_window_bits", "10")],
                [("server_max_window_bits", "10")],
                (False, False, 15, 10),
            ),
            (
                (False, False, 12, None),
                [("server_max_window_bits", "12")],
                [("server_max_window_bits", "12")],
                (False, False, 15, 12),
            ),
            (
                (False, False, 12, None),
                [("server_max_window_bits", "13")],
                [("server_max_window_bits", "12")],
                (False, False, 15, 12),
            ),
            (
                (False, False, 12, None),
                [("server_max_window_bits", "12")] * 2,
                None,
                DuplicateParameter,
            ),
            (
                (False, False, 12, None),
                [("server_max_window_bits", "42")],
                None,
                InvalidParameterValue,
            ),
            # Test client_max_window_bits
            (
                (False, False, None, None),
                [("client_max_window_bits", None)],
                [],
                (False, False, 15, 15),
            ),
            (
                (False, False, None, None),
                [("client_max_window_bits", "7")],
                None,
                InvalidParameterValue,
            ),
            (
                (False, False, None, None),
                [("client_max_window_bits", "10")],
                [("client_max_window_bits", "10")],  # doesn't matter
                (False, False, 10, 15),
            ),
            (
                (False, False, None, None),
                [("client_max_window_bits", "16")],
                None,
                InvalidParameterValue,
            ),
            (
                (False, False, None, 12),
                [],
                [],
                (False, False, 15, 15),
            ),
            (
                (False, False, None, 12, {}, True),
                [],
                None,
                NegotiationError,
            ),
            (
                (False, False, None, 12),
                [("client_max_window_bits", None)],
                [("client_max_window_bits", "12")],
                (False, False, 12, 15),
            ),
            (
                (False, False, None, 12),
                [("client_max_window_bits", "10")],
                [("client_max_window_bits", "10")],
                (False, False, 10, 15),
            ),
            (
                (False, False, None, 12),
                [("client_max_window_bits", "12")],
                [("client_max_window_bits", "12")],  # doesn't matter
                (False, False, 12, 15),
            ),
            (
                (False, False, None, 12),
                [("client_max_window_bits", "13")],
                [("client_max_window_bits", "12")],  # doesn't matter
                (False, False, 12, 15),
            ),
            (
                (False, False, None, 12),
                [("client_max_window_bits", "12")] * 2,
                None,
                DuplicateParameter,
            ),
            (
                (False, False, None, 12),
                [("client_max_window_bits", "42")],
                None,
                InvalidParameterValue,
            ),
            # Test all parameters together
            (
                (True, True, 12, 12),
                [
                    ("server_no_context_takeover", None),
                    ("client_no_context_takeover", None),
                    ("server_max_window_bits", "10"),
                    ("client_max_window_bits", "10"),
                ],
                [
                    ("server_no_context_takeover", None),
                    ("client_no_context_takeover", None),
                    ("server_max_window_bits", "10"),
                    ("client_max_window_bits", "10"),
                ],
                (True, True, 10, 10),
            ),
            (
                (False, False, None, None),
                [
                    ("server_no_context_takeover", None),
                    ("client_no_context_takeover", None),
                    ("server_max_window_bits", "10"),
                    ("client_max_window_bits", "10"),
                ],
                [
                    ("server_no_context_takeover", None),
                    ("client_no_context_takeover", None),
                    ("server_max_window_bits", "10"),
                    ("client_max_window_bits", "10"),
                ],
                (True, True, 10, 10),
            ),
            (
                (True, True, 12, 12),
                [("client_max_window_bits", None)],
                [
                    ("server_no_context_takeover", None),
                    ("client_no_context_takeover", None),
                    ("server_max_window_bits", "12"),
                    ("client_max_window_bits", "12"),
                ],
                (True, True, 12, 12),
            ),
        ]:
            with self.subTest(
                config=config,
                request_params=request_params,
                response_params=response_params,
            ):
                factory = ServerPerMessageDeflateFactory(*config)
                if isinstance(result, type) and issubclass(result, Exception):
                    with self.assertRaises(result):
                        factory.process_request_params(request_params, [])
                else:
                    params, extension = factory.process_request_params(
                        request_params, []
                    )
                    self.assertEqual(params, response_params)
                    expected = PerMessageDeflate(*result)
                    self.assertExtensionEqual(extension, expected)

    def test_process_response_params_deduplication(self):
        factory = ServerPerMessageDeflateFactory(False, False, None, None)
        with self.assertRaises(NegotiationError):
            factory.process_request_params(
                [], [PerMessageDeflate(False, False, 15, 15)]
            )

    def test_enable_server_permessage_deflate(self):
        for extensions, (
            expected_len,
            expected_position,
            expected_compress_settings,
        ) in [
            (
                None,
                (1, 0, {"memLevel": 5}),
            ),
            (
                [],
                (1, 0, {"memLevel": 5}),
            ),
            (
                [ServerNoOpExtensionFactory()],
                (2, 1, {"memLevel": 5}),
            ),
            (
                [ServerPerMessageDeflateFactory(compress_settings={"memLevel": 7})],
                (1, 0, {"memLevel": 7}),
            ),
            (
                [
                    ServerPerMessageDeflateFactory(compress_settings={"memLevel": 7}),
                    ServerNoOpExtensionFactory(),
                ],
                (2, 0, {"memLevel": 7}),
            ),
            (
                [
                    ServerNoOpExtensionFactory(),
                    ServerPerMessageDeflateFactory(compress_settings={"memLevel": 7}),
                ],
                (2, 1, {"memLevel": 7}),
            ),
        ]:
            with self.subTest(extensions=extensions):
                extensions = enable_server_permessage_deflate(extensions)
                self.assertEqual(len(extensions), expected_len)
                extension = extensions[expected_position]
                self.assertIsInstance(extension, ServerPerMessageDeflateFactory)
                self.assertEqual(
                    extension.compress_settings,
                    expected_compress_settings,
                )



import asyncio
import contextlib

from websockets.asyncio.connection import Connection


class InterceptingConnection(Connection):
    """
    Connection subclass that can intercept outgoing packets.

    By interfacing with this connection, we simulate network conditions
    affecting what the component being tested receives during a test.

    """

    def connection_made(self, transport):
        super().connection_made(InterceptingTransport(transport))

    @contextlib.contextmanager
    def delay_frames_sent(self, delay):
        """
        Add a delay before sending frames.

        This can result in out-of-order writes, which is unrealistic.

        """
        assert self.transport.delay_write is None
        self.transport.delay_write = delay
        try:
            yield
        finally:
            self.transport.delay_write = None

    @contextlib.contextmanager
    def delay_eof_sent(self, delay):
        """
        Add a delay before sending EOF.

        This can result in out-of-order writes, which is unrealistic.

        """
        assert self.transport.delay_write_eof is None
        self.transport.delay_write_eof = delay
        try:
            yield
        finally:
            self.transport.delay_write_eof = None

    @contextlib.contextmanager
    def drop_frames_sent(self):
        """
        Prevent frames from being sent.

        Since TCP is reliable, sending frames or EOF afterwards is unrealistic.

        """
        assert not self.transport.drop_write
        self.transport.drop_write = True
        try:
            yield
        finally:
            self.transport.drop_write = False

    @contextlib.contextmanager
    def drop_eof_sent(self):
        """
        Prevent EOF from being sent.

        Since TCP is reliable, sending frames or EOF afterwards is unrealistic.

        """
        assert not self.transport.drop_write_eof
        self.transport.drop_write_eof = True
        try:
            yield
        finally:
            self.transport.drop_write_eof = False


class InterceptingTransport:
    """
    Transport wrapper that intercepts calls to ``write()`` and ``write_eof()``.

    This is coupled to the implementation, which relies on these two methods.

    Since ``write()`` and ``write_eof()`` are not coroutines, this effect is
    achieved by scheduling writes at a later time, after the methods return.
    This can easily result in out-of-order writes, which is unrealistic.

    """

    def __init__(self, transport):
        self.loop = asyncio.get_running_loop()
        self.transport = transport
        self.delay_write = None
        self.delay_write_eof = None
        self.drop_write = False
        self.drop_write_eof = False

    def __getattr__(self, name):
        return getattr(self.transport, name)

    def write(self, data):
        if not self.drop_write:
            if self.delay_write is not None:
                self.loop.call_later(self.delay_write, self.transport.write, data)
            else:
                self.transport.write(data)

    def write_eof(self):
        if not self.drop_write_eof:
            if self.delay_write_eof is not None:
                self.loop.call_later(self.delay_write_eof, self.transport.write_eof)
            else:
                self.transport.write_eof()



import asyncio
import json
import logging

from websockets.asyncio.client import connect
from websockets.exceptions import WebSocketException


logging.basicConfig(level=logging.WARNING)

SERVER = "ws://localhost:9001"

AGENT = "websockets.asyncio"


async def get_case_count():
    async with connect(f"{SERVER}/getCaseCount") as ws:
        return json.loads(await ws.recv())


async def run_case(case):
    async with connect(
        f"{SERVER}/runCase?case={case}&agent={AGENT}",
        max_size=2**25,
    ) as ws:
        try:
            async for msg in ws:
                await ws.send(msg)
        except WebSocketException:
            pass


async def update_reports():
    async with connect(
        f"{SERVER}/updateReports?agent={AGENT}",
        open_timeout=60,
    ):
        pass


async def main():
    cases = await get_case_count()
    for case in range(1, cases + 1):
        print(f"Running test case {case:03d} / {cases}... ", end="\t")
        try:
            await run_case(case)
        except WebSocketException as exc:
            print(f"ERROR: {type(exc).__name__}: {exc}")
        except Exception as exc:
            print(f"FAIL: {type(exc).__name__}: {exc}")
        else:
            print("OK")
    print(f"Ran {cases} test cases")
    await update_reports()
    print("Updated reports")


if __name__ == "__main__":
    asyncio.run(main())
